"""
  Unit tests for gp.py
  -- kandasamy@cs.cmu.edu
"""
from __future__ import absolute_import
from __future__ import division

# pylint: disable=import-error
# pylint: disable=no-member
# pylint: disable=invalid-name
# pylint: disable=relative-import

from ..utils.base_test_class import BaseTestClass, execute_tests
import numpy as np
# Local
from .gp_core import GP
from .kernel import SEKernel


def gen_gp_test_data():
  """ This function generates a bunch of test data. """
  # pylint: disable=too-many-locals
  # Dataset 1
  f1 = lambda x: (x**2).sum(axis=1)
  N1 = 5
  X1_tr = np.array(list(range(N1))).astype(float).reshape((N1, 1))/N1 + 1/(2*N1)
  Y1_tr = f1(X1_tr)
  X1_te = np.random.random((50, 1))
  Y1_te = f1(X1_te)
  kernel1 = SEKernel(1, 1, 0.5)
  # Dataset 2
  N2 = 100
  D2 = 10
  f2 = lambda x: ((x**2) * list(range(1, D2+1))/D2).sum(axis=1)
  X2_tr = np.random.random((N2, D2))
  Y2_tr = f2(X2_tr)
  X2_te = np.random.random((N2, D2))
  Y2_te = f2(X2_te)
  kernel2 = SEKernel(D2, 10, 0.2*np.sqrt(D2))
  # Dataset 3
  N3 = 200
  D3 = 6
  f3 = lambda x: ((x**3 + 2 * x**2 - x + 2) * list(range(1, D3+1))/D3).sum(axis=1)
  X3_tr = np.random.random((N3, D3))
  Y3_tr = f3(X3_tr)
  X3_te = np.random.random((N3, D3))
  Y3_te = f3(X3_te)
  kernel3 = SEKernel(D3, 10, 0.2*np.sqrt(D3))
  # Dataset 4
  N4 = 400
  D4 = 8
  f4 = lambda x: ((np.sin(x**2) + 2 * np.cos(x**2) - x + 2) *
                  list(range(1, D4+1))/D4).sum(axis=1)
  X4_tr = np.random.random((N4, D4))
  Y4_tr = f4(X4_tr)
  X4_te = np.random.random((N4, D4))
  Y4_te = f4(X4_te)
  kernel4 = SEKernel(D4, 10, 0.2*np.sqrt(D4))
  # put all datasets into a list.
  return [(X1_tr, Y1_tr, kernel1, X1_te, Y1_te),
          (X2_tr, Y2_tr, kernel2, X2_te, Y2_te),
          (X3_tr, Y3_tr, kernel3, X3_te, Y3_te),
          (X4_tr, Y4_tr, kernel4, X4_te, Y4_te)]


def build_gp_with_dataset(dataset, kern=None):
  """ Function to build a SE GP with the dataset. """
  mean_func = lambda x: np.array([np.median(dataset[1])] * len(x))
  noise_var = dataset[1].std()**2/20
  kern = dataset[2] if kern is None else kern
  return GP(dataset[0], dataset[1], kern, mean_func, noise_var)


def compute_average_prediction_error(dataset, preds):
  """ Computes the prediction error. """
  return (np.linalg.norm(dataset[4] - preds)**2)/len(dataset[4])


class GPTestCase(BaseTestClass):
  """ Unit tests for the GP class. """
  # pylint: disable=too-many-locals

  def setUp(self):
    """ Set up for tests. """
    self.datasets = gen_gp_test_data()

  def test_add_data(self):
    """ Tests GP.add_data. """
    self.report('GP.add_data')
    for dataset in self.datasets:
      num_new = np.random.randint(3, 10)
      X_new = np.random.random((num_new, dataset[0].shape[1]))
      Y_new = np.random.random(num_new)
      curr_gp = build_gp_with_dataset(dataset)
      curr_gp.add_data_multiple(X_new, Y_new)
      assert num_new + len(dataset[1]) == curr_gp.num_tr_data
      x_new = np.random.random(dataset[0].shape[1])
      y_new = np.random.random()
      curr_gp.add_data_single(x_new, y_new)
      assert 1 + num_new + len(dataset[1]) == curr_gp.num_tr_data

  def test_eval(self):
    """ Tests the evaluation. """
    self.report('GP.eval: Probabilistic test, might fail sometimes')
    num_successes = 0
    for dataset in self.datasets:
      curr_gp = build_gp_with_dataset(dataset)
      curr_pred, _ = curr_gp.eval(dataset[3])
      curr_err = compute_average_prediction_error(dataset, curr_pred)
      const_err = compute_average_prediction_error(dataset, dataset[1].mean())
      success = curr_err < const_err
      self.report(('(N,D)=' + str(dataset[0].shape) + ':: GP-err= ' + str(curr_err) +
                   ',   Const-err= ' + str(const_err) + ',  success=' + str(success)),
                  'test_result')
      num_successes += int(success)
    assert num_successes > 0.6 *len(self.datasets)

  def test_compute_log_marginal_likelihood(self):
    """ Tests compute_log_marginal_likelihood. Does not test for accurate implementation.
        Only tests if the function runs without runtime errors. """
    self.report('GP.compute_log_marginal_likelihood: ** Runtime test errors only **')
    for dataset in self.datasets:
      curr_gp = build_gp_with_dataset(dataset)
      lml = curr_gp.compute_log_marginal_likelihood()
      self.report('(N,D)=' + str(dataset[0].shape) + ' lml = ' + str(lml),
                  'test_result')


if __name__ == '__main__':
  execute_tests()




"""
  Harness for storing and computing various representations for neural networks.
  --kandasamy@cs.cmu.edu
"""

# pylint: disable=no-member
# pylint: disable=invalid-name
# pylint: disable=abstract-class-not-used

from copy import deepcopy
import numpy as np
# Local imports
from ..utils import graph_utils
from ..utils.general_utils import get_nonzero_indices_in_vector, reorder_list_or_array, \
                                  reorder_rows_and_cols_in_matrix

# global constants
STORE_GRAPH_PARAMS_BY_DEFAULT = False
SOFTMAX_LINEAR_LAYER_MASS = 0
POOL_LAYER_MASS = 0
UNIV_MLP_RECTIFIERS = ['relu', 'relu6', 'crelu', 'relu-x', 'leaky-relu', 'softplus',
                       'elu']
UNIV_MLP_SIGMOIDS = ['logistic', 'tanh', 'step']
MLP_RECTIFIERS = ['relu', 'crelu', 'leaky-relu', 'softplus', 'elu']
MLP_SIGMOIDS = ['logistic', 'tanh']
_NON_PROC_LAYERY_MASS_FRAC = 0.1
_FC_LAYER_MASS_COEFF = 0.1 # coefficient for the fully connected layers


# Some utilities we will need below ---------------------------------------------
def is_a_pooling_layer_label(layer_label):
  """ Returns true if a pooling layer. """
  return 'pool' in layer_label

def is_a_conv_layer_label(layer_label):
  """ Returns true if a convolutional layer. """
  return 'conv' in layer_label or 'res' in layer_label


# Exceptions we will define that can be handled later on ------------------------
class CNNImageSizeMismatchException(Exception):
  """ Exception for mismatches in image sizes. """
  def __init__(self, descr):
    """ Constructor. """
    super(CNNImageSizeMismatchException, self).__init__()
    self.descr = descr
  def __str__(self):
    """ Returns string representation. """
    return 'CNNImageSizeMismatchException: %s'%(self.descr)

class CNNNoConvAfterIPException(Exception):
  """ Exception if the layers immediately after the input are not convloutions. """
  def __init__(self, descr):
    """ Constructor. """
    super(CNNNoConvAfterIPException, self).__init__()
    self.descr = descr
  def __str__(self):
    """ Returns string representation. """
    return 'CNNNoConvAfterIPException: %s'%(self.descr)


# Neural network class ----------------------------------------------------------
class NeuralNetwork(object):
  """ Base class for all neural networks. """
  # pylint: disable=attribute-defined-outside-init

  def __init__(self, nn_class, layer_labels, conn_mat, num_units_in_each_layer,
               all_layer_label_classes, layer_label_similarities=None):
    """ nn_class is a string describing whether it is a FF network, CNN etc.
      Parameters specific to the NN
      layer labels: is a list of labels for each layer.
      conn_mat: is the connectivity matrix. If (i,j) is 1 it means there is a
                connection between layer i and j.
      Global parameters
      all_layer_label_classes: a list of all labels a network can take.
      layer_label_similarities: measures of similarity for each label type.
    """
    self.nn_class = nn_class
    self.layer_labels = layer_labels
    self.conn_mat = conn_mat
    self.num_units_in_each_layer = np.array(num_units_in_each_layer)
    if not hasattr(self, 'mandatory_child_attributes'):
      self.mandatory_child_attributes = []
    # Global parameters
    self.all_layer_label_classes = all_layer_label_classes
    self.layer_label_similarities = layer_label_similarities
    self._set_up()

  def _set_up(self):
    """ Additional constructor actions. """
    # Prelims
    self.num_layers = len(self.layer_labels)
    self.bkwd_ip_dists_of_layers = None
    self.fwd_op_dists_of_layers = None
    # First do a topological sort
    self._topological_sort()
    # Compute layer masses
    self._compute_layer_masses()
    # Check for other parameters
    self.num_internal_layers = len(self.layer_labels) - 2
    self.num_processing_layers = len([ll for ll in self.layer_labels if ll not in
                                      ['ip', 'op', 'softmax', 'linear']])
    self.internal_layer_idxs = np.delete(range(self.num_layers), [self.get_ip_layer_idx(),
                                 self.get_op_layer_idx()])
    self.internal_layer_masses = self.layer_masses[self.internal_layer_idxs]
    # For storing additional matrices we might need along the way
    # Compute the distances by short circuiting each type of layer.
    self._bkwd_ip_fwd_op_dist_type_order = ['all'] + self._get_child_layer_groups()
    self._path_length_type_order = ['shortest', 'longest', 'rw']
    self._all_path_length_categories = [x + '-' + y
                                        for x in self._bkwd_ip_fwd_op_dist_type_order
                                        for y in self._path_length_type_order]
    self._compute_ip_op_path_lengths()
    # Now re-organise the above data into a matrix of num-layer x num-layer-type matrix.
    self.bkwd_ip_dists_of_layers, self.fwd_op_dists_of_layers = \
      self.get_bkwd_ip_fwd_op_dists_of_all_layers(self._all_path_length_categories)
    # Finally, check if this is a valid network
    assert self._check_if_valid_network()

  def _topological_sort(self):
    """ Reorders the layers so that they are in a topologically sorted fashion. """
    top_order, has_cycles = graph_utils.kahn_topological_sort(self.conn_mat,
                              self.layer_labels.index('ip'))
    assert not has_cycles
    self.layer_labels = reorder_list_or_array(self.layer_labels, top_order)
    self.num_units_in_each_layer = reorder_list_or_array(self.num_units_in_each_layer,
                                                      top_order)
    self.conn_mat = reorder_rows_and_cols_in_matrix(self.conn_mat, top_order)
    self._child_attrs_topological_sort(top_order)

  def _compute_layer_masses(self):
    """ Computes the layer masses for all layers. """
    self._child_compute_layer_masses()

  def _child_compute_layer_masses(self):
    """ Computes the layer masses for all layers. """
    raise NotImplementedError('Implement _child_compute_layer_masses in a child class.')

  def get_children(self, layer_idx):
    """ Returns the children of layer_idx. """
    return graph_utils.get_children(layer_idx, self.conn_mat)

  def get_parents(self, layer_idx):
    """ Returns the parents of layer_idx. """
    return graph_utils.get_parents(layer_idx, self.conn_mat)

  def get_edges(self):
    """ Returns the edges of network as a list. """
    return self.conn_mat.keys()

  def get_total_num_edges(self):
    """ Returns the total number of edges. """
    return self.conn_mat.sum()

  def _get_layer_indices_of_layer_type(self, layer_type, layer_labels=None):
    """ Gets the layer indices of layer_type. """
    return self._get_layer_indices_of_layer_or_group_type(layer_type, 'layer',
                                                          layer_labels)

  def _get_layer_indices_of_group_type(self, group_type, layer_labels=None):
    """ Gets the layer indices of layers belonging to that group. """
    return self._get_layer_indices_of_layer_or_group_type(group_type, 'group',
                                                          layer_labels)

  def _get_layer_indices_of_layer_or_group_type(self, label, layer_or_group=None,
                                                layer_labels=None):
    """ Gets the layer indices of layers belonging to the group 'label' if layer_or_group
        is 'group'. Else it returns the indices where the layer_label is label. """
    layer_labels = layer_labels if layer_labels is not None else self.layer_labels
    if label == 'all':
      return list(range(len(layer_labels)))
    # Otherwise first determine whether it is a group or a layer
    if layer_or_group == None:
      group_label = self._get_layer_group_for_layer_label(label)
      layer_or_group = 'group' if group_label is None else 'layer'
    # Now return accordingly
    if layer_or_group == 'layer':
      return [i for i, x in enumerate(layer_labels) if x == label]
    elif layer_or_group == 'group':
      return [i for i, x in enumerate(layer_labels) if
              self._get_layer_group_for_layer_label(x) == label]
    else:
      raise ValueError('layer_or_group should be either \'layer\' or \'group\' or None.')

  def get_ip_layer_idx(self):
    """ Returns the index of the input layer. """
    return self.layer_labels.index('ip')

  def get_op_layer_idx(self):
    """ Returns the index of the output layer. """
    return self.layer_labels.index('op')

  def get_total_mass(self):
    """ Returns the total mass of the network. """
    return self.layer_masses.sum()

  def get_in_degrees(self):
    """ Returns the in degrees of all nodes. """
    return np.array(self.conn_mat.sum(axis=0)).ravel()

  def get_out_degrees(self):
    """ Returns the out degrees of all nodes. """
    return np.array(self.conn_mat.sum(axis=1)).ravel()

  def get_distances_from_ip(self, dist_type='all-shortest'):
    """ Returns the distances from the input. """
    return self._bkwd_dists_to_ip[dist_type]

  def get_distances_to_op(self, dist_type='all-shortest'):
    """ Returns the distances from the input. """
    return self._fwd_dists_to_op[dist_type]

  def _check_if_valid_network(self):
    """ Returns true if the network is a valid network. """
    # Check if the lenghts of the labels and node masss are the same
    assert len(self.layer_labels) == len(self.num_units_in_each_layer)
    # If there are no processing layers, there need not be more than 3 layers (ip, op and
    # decision).
    assert self.num_processing_layers > 0 or self.num_layers == 3
    # Check if the labels are correct
    input_layers = self._get_layer_indices_of_layer_type('ip')
    output_layers = self._get_layer_indices_of_layer_type('op')
    assert len(input_layers) == 1
    assert len(output_layers) == 1
    # A bunch of checks based on the APSP distances
    # Check if the forward distances to 'ip' and backward distances to 'op' are 0.
    ip_idx = self.get_ip_layer_idx()
    op_idx = self.get_op_layer_idx()
    # Check if self distances are all 0
    assert np.all(self._fwd_dists_to_op['all-rw'][op_idx] == 0)
    assert np.all(self._bkwd_dists_to_ip['all-rw'][ip_idx] == 0)
    # Check to see if the forward distances from the ip and backward distances from the
    # output are finite.
    assert np.all(np.isfinite(self._fwd_dists_to_op['all-longest']))
    assert np.all(np.isfinite(self._bkwd_dists_to_ip['all-longest']))
    # Check if all labels are valid
    assert all([ll in self.all_layer_label_classes for ll in self.layer_labels])
    # Finally check if there are any conditions that the child network needs to satisfy.
    assert self._child_check_if_valid_network()
    return True

  def _child_check_if_valid_network(self):
    """ Used for any tests the child class might want to check. """
    raise NotImplementedError('Impelement in a child class.')

  def _child_attrs_topological_sort(self, top_order):
    """ Topologically sort any attributes of the child class. """
    raise NotImplementedError('Impelement in a child class.')

  def get_layer_descr(self, layer_idx, *_):
    """ Returns a string describing the layer. Used in visualing the layer. """
    if isinstance(self.num_units_in_each_layer[layer_idx], (int, float)) and \
      np.isfinite(self.num_units_in_each_layer[layer_idx]):
      num_units_descr = str(int(self.num_units_in_each_layer[layer_idx])) + ','
    else:
      num_units_descr = ''
    return '#%d %s, %s\n(%d)'%(layer_idx, self.layer_labels[layer_idx],
                             num_units_descr, self.layer_masses[layer_idx])
#     return '#%d %s, %s'%(layer_idx, self.layer_labels[layer_idx],
#                          num_units_descr) # cleantech slides

  def get_edge_weights_from_conn_mat(self):
    """ Returns a matrix of edge weights with infinities for non-edges. """
    # First compute the preliminary edge weights matrix.
    edge_weights = (deepcopy(self.conn_mat)).toarray()
    edge_weights[edge_weights == 0] = np.inf
    return edge_weights

  def get_layer_or_group_edge_weights_from_edge_weights(self, edge_weights,
                                                        layer_or_group_label):
    """ Gets the edge weights corresponding to the layer_or_group_label. """
    layer_or_group_edge_weights = deepcopy(edge_weights)
    layer_or_group_edge_weights[layer_or_group_edge_weights == 1] = 0
    curr_layers = self._get_layer_indices_of_layer_or_group_type(layer_or_group_label)
    for cl in curr_layers:
      layer_or_group_edge_weights[cl, :] = edge_weights[cl, :]
    return layer_or_group_edge_weights

  def _compute_apsps(self):
    """ Computes various kinds of all pairs shortest paths. """
    edge_weights = self.get_edge_weights_from_conn_mat()
    self._fwd_apsps = {}
    for lg in self._bkwd_ip_fwd_op_dist_type_order:
      curr_edge_weights = self.get_layer_or_group_edge_weights_from_edge_weights(
                            edge_weights, lg)
      self._fwd_apsps[lg] = graph_utils.apsp_floyd_warshall_costs(curr_edge_weights)

  def _compute_ip_op_path_lengths(self):
    """ Computes the various kinds of shortest path lengths to ip and op. """
    edge_weights = self.get_edge_weights_from_conn_mat()
    self._fwd_dists_to_op = {}
    self._bkwd_dists_to_ip = {}
    for lg in self._bkwd_ip_fwd_op_dist_type_order:
      curr_edge_weights = self.get_layer_or_group_edge_weights_from_edge_weights(
                            edge_weights, lg)
      curr_edge_weights_T = self.get_layer_or_group_edge_weights_from_edge_weights(
                              edge_weights.T, lg)
      for plt in self._path_length_type_order:
        curr_key = lg + '-' + plt
        # compute_nn_path_lengths expects the forward graph and computes the distances
        # to the last node.
        self._bkwd_dists_to_ip[curr_key] = graph_utils.compute_nn_path_lengths(
          curr_edge_weights_T, list(reversed(range(self.num_layers))), plt)
        self._fwd_dists_to_op[curr_key] = graph_utils.compute_nn_path_lengths(
          curr_edge_weights, list((range(self.num_layers))), plt)

  def get_bkwd_ip_fwd_op_dists_of_layer(self, layer_idx, path_length_types):
    """ Returns the forward and backward distances of the layer in layer_idx. """
    bkwd_ip_dists = [self._bkwd_dists_to_ip[dist_type][layer_idx] for dist_type in
                     path_length_types]
    fwd_op_dists = [self._fwd_dists_to_op[dist_type][layer_idx] for dist_type in
                    path_length_types]
    return bkwd_ip_dists, fwd_op_dists

  def get_bkwd_ip_fwd_op_dists_of_all_layers(self, path_length_categories):
    """ Returns the forward and backward distances of all layers except the input/output
        layers. """
    if self.bkwd_ip_dists_of_layers is not None:
      col_reordering = [self._all_path_length_categories.index(elem)
                        for elem in path_length_categories]
      return (self.bkwd_ip_dists_of_layers[:, col_reordering],
              self.fwd_op_dists_of_layers[:, col_reordering])
    else:
      bkwd_ip_dists_of_layers = []
      fwd_op_dists_of_layers = []
      for lidx in range(self.num_layers):
        curr_bip, curr_fop = self.get_bkwd_ip_fwd_op_dists_of_layer(lidx,
                               path_length_categories)
        bkwd_ip_dists_of_layers.append(curr_bip)
        fwd_op_dists_of_layers.append(curr_fop)
      return np.array(bkwd_ip_dists_of_layers), np.array(fwd_op_dists_of_layers)

  def _get_property_and_set_if_to_store(self, property_name, property_comp_function,
                                        to_store=STORE_GRAPH_PARAMS_BY_DEFAULT):
    """ A function for returning and setting a property if required. Will be used by many
        of the functions below. """
    curr_stored_val = getattr(self, property_name)
    if curr_stored_val is not None:
      return curr_stored_val
    else:
      computed_val = property_comp_function()
      if to_store:
        setattr(self, property_name, computed_val)
      return computed_val

  # Some abstract methods ============================================================
  @classmethod
  def _get_child_layer_groups(cls):
    """ Returns layer groups for the network. """
    raise NotImplementedError('Implement for a child class.')

  @classmethod
  def _get_layer_group_for_layer_label(cls, layer_label):
    """ Returns layer group for layer. """
    raise NotImplementedError('Implement for a child class.')



# Some specific instances of Neural network ============================================
def _check_if_layers_before_op_are(conn_mat, op_layer_idx, layer_labels, label_val):
  """ Checks if all layers before the output have label_val as the label. """
  layers_before_op = get_nonzero_indices_in_vector(conn_mat[:, op_layer_idx])
  for layer_idx in layers_before_op:
    if layer_labels[layer_idx] != label_val:
      return False
  return True

def compute_num_channels_at_each_layer(nn):
  """ Computes the number of channels at each layer. """
  num_channels_in_to_each_layer = [None]
  num_channels_out_of_each_layer = [1]
  for layer_idx in range(1, nn.num_layers-1):
    ll = nn.layer_labels[layer_idx]
    curr_parents = nn.get_parents(layer_idx)
    channels_from_parents = [num_channels_out_of_each_layer[par_idx] for
                             par_idx in curr_parents]
    # Determine the number of channels in
    curr_in = sum(channels_from_parents)
    # Determine the number of channels out
    if is_a_pooling_layer_label(ll):
      curr_out = curr_in
    elif ll in ['softmax', 'linear']:
      curr_out = None
    else:
      curr_out = nn.num_units_in_each_layer[layer_idx]
    num_channels_in_to_each_layer.append(curr_in)
    num_channels_out_of_each_layer.append(curr_out)
  # For the output layer
  num_channels_in_to_each_layer.append(None)
  num_channels_out_of_each_layer.append(None)
  return num_channels_in_to_each_layer, num_channels_out_of_each_layer

def compute_layer_masses(num_channels_in_to_each_layer, num_channels_out_of_each_layer,
                         layer_labels):
  """ Computes the layer masses for all layers. """
  num_layers = len(num_channels_in_to_each_layer)
  layer_masses = np.zeros((num_layers))
  num_decision_layers = 0
  for layer_idx in range(num_layers):
    if layer_labels[layer_idx] in ['softmax', 'linear']:
      num_decision_layers += 1
      continue # we will do this later.
    elif layer_labels[layer_idx] in ['ip', 'op']:
      continue # we will do this later.
    if is_a_pooling_layer_label(layer_labels[layer_idx]):
      layer_masses[layer_idx] = num_channels_in_to_each_layer[layer_idx]
    else:
      layer_masses[layer_idx] = \
        num_channels_in_to_each_layer[layer_idx] * \
        num_channels_out_of_each_layer[layer_idx]
      if layer_labels[layer_idx] == 'fc':
        layer_masses[layer_idx] *= _FC_LAYER_MASS_COEFF
  # Now add layer mass for the non-processing layers
  total_proc_layer_mass = layer_masses.sum()
  non_proc_layer_mass = max(_NON_PROC_LAYERY_MASS_FRAC * total_proc_layer_mass, 100)
  decision_layer_mass = non_proc_layer_mass / float(num_decision_layers)
  for layer_idx, ll in enumerate(layer_labels):
    if ll in ['softmax', 'linear']:
      layer_masses[layer_idx] = decision_layer_mass
    elif ll in ['ip', 'op']:
      layer_masses[layer_idx] = non_proc_layer_mass
  return layer_masses


class ConvNeuralNetwork(NeuralNetwork):
  """ Implements a convolutional neural network. """
  # pylint: disable=attribute-defined-outside-init

  def __init__(self, layer_labels, conn_mat, num_filters_in_each_layer,
               strides, all_layer_label_classes=None, layer_label_similarities=None):
    """ Constructor. """
    self.mandatory_child_attributes = ['strides']
    self.strides = strides
    super(ConvNeuralNetwork, self).__init__('cnn', layer_labels, conn_mat,
          num_filters_in_each_layer, all_layer_label_classes,
          layer_label_similarities)

  def _child_compute_layer_masses(self):
    """ Computes the layer masses and the number of channels at each layer. """
    self.num_channels_in_to_each_layer, self.num_channels_out_of_each_layer = \
      compute_num_channels_at_each_layer(self)
    self.layer_masses = compute_layer_masses(self.num_channels_in_to_each_layer,
      self.num_channels_out_of_each_layer, self.layer_labels)

  @classmethod
  def _check_if_parent_image_sizes_are_valid(cls, image_sizes):
    """ Returns true if all image sizes are equal. """
    if len(image_sizes) == 0:
      return 'This node has no parents.'
    elif not image_sizes[1:] == image_sizes[:-1]:
      return 'Not all image sizes are equal: %s.'%(str(image_sizes))
    elif image_sizes[0] == None:
      return 'Image size None encountered: %s.'%(str(image_sizes))
    return ''

  def _check_image_size_consistency(self):
    """ Checks if the images sizes are consistent across all layers. """
    self.pre_img_inv_sizes = [None for _ in range(self.num_layers)]
    self.post_img_inv_sizes = [None for _ in range(self.num_layers)]
    self.pre_img_inv_sizes[0] = 1
    self.post_img_inv_sizes[0] = 1
    assert self.strides[0] is None
    for layer_idx in range(1, self.num_layers):
      # First check if the stride value is consistent
      if is_a_conv_layer_label(self.layer_labels[layer_idx]):
        assert self.strides[layer_idx] in [1, 2]
      else:
        assert self.strides[layer_idx] is None
      # Obtain the parents and check if the sizes are consistent
      curr_parents = self.get_parents(layer_idx)
      parent_post_img_sizes = [self.post_img_inv_sizes[x] for x in curr_parents]
      # Check that the sizes are valid if it is a convolutional or pooling layer.
      if is_a_pooling_layer_label(self.layer_labels[layer_idx]) or \
         is_a_conv_layer_label(self.layer_labels[layer_idx]):
        # Checking only the first parent because we are checking for equality later.
        assert isinstance(parent_post_img_sizes[0], (int, float)) and \
               parent_post_img_sizes[0] > 0
      # Check parent image sizes and if they are consistent
      check_parent_img_sizes = self._check_if_parent_image_sizes_are_valid(
                                 parent_post_img_sizes)
      if check_parent_img_sizes != '':
        raise CNNImageSizeMismatchException('layer %d (%s): %s.'%(layer_idx,
                          self.layer_labels[layer_idx], check_parent_img_sizes))
      # If so, now update the child's parameters
      self.pre_img_inv_sizes[layer_idx] = parent_post_img_sizes[0]
      if is_a_pooling_layer_label(self.layer_labels[layer_idx]) or \
          (is_a_conv_layer_label(self.layer_labels[layer_idx]) and
           self.strides[layer_idx] == 2):
        self.post_img_inv_sizes[layer_idx] = 2 * self.pre_img_inv_sizes[layer_idx]
      elif is_a_conv_layer_label(self.layer_labels[layer_idx]):
        self.post_img_inv_sizes[layer_idx] = self.pre_img_inv_sizes[layer_idx]
      elif self.layer_labels[layer_idx] in ['fc', 'softmax', 'op']:
        self.post_img_inv_sizes[layer_idx] = 'x'
    return True

  def _check_layer_units_and_masses(self):
    """ Checks the number of units and masses for the pool layers. """
    for idx, ll in enumerate(self.layer_labels):
      if is_a_conv_layer_label(ll):
        assert self.num_units_in_each_layer[idx] > 0
        assert self.layer_masses[idx] > 0
      elif is_a_pooling_layer_label(ll):
        pass
      elif ll in ['ip', 'op', 'softmax', 'linear']:
        assert self.num_units_in_each_layer[idx] is None
    return True

  def _check_layers_after_ip(self):
    """ Checks if layers after the input are all convolutional layers. """
    ip_ch_labels = [self.layer_labels[i] for i in self.get_children(0)]
    ip_ch_labels_are_conv = [is_a_conv_layer_label(lab) for lab in ip_ch_labels]
    if not all(ip_ch_labels_are_conv):
      raise CNNNoConvAfterIPException('Children of input layer are not convolutional:' +
        ' %s.'%(ip_ch_labels))
    return True

  def _child_check_if_valid_network(self):
    """ Runs any checks relevant for a CNN. """
    # Test if all layers before the output are softmax.
    assert _check_if_layers_before_op_are(self.conn_mat, self.get_op_layer_idx(),
                                          self.layer_labels, 'softmax')
    assert self._check_image_size_consistency()
    assert self._check_layer_units_and_masses()
    assert self._check_layers_after_ip()
    return True

  def _child_attrs_topological_sort(self, top_order):
    """ Topologically sort child attributes. """
    self.strides = reorder_list_or_array(self.strides, top_order)

  def get_layer_descr(self, layer_idx, for_pres=False):
    """ Returns a string description describing the layer. """
    #pylint: disable=arguments-differ
    if not for_pres:
      if is_a_conv_layer_label(self.layer_labels[layer_idx]) and \
         self.strides[layer_idx] == 2:
        stride_str = ', /2'
      else:
        stride_str = ''
      img_size_str = 'None' if self.pre_img_inv_sizes[layer_idx] is None else \
                     str(self.pre_img_inv_sizes[layer_idx])
      child_str = ' [%s%s]'%(img_size_str, stride_str)
      return super(ConvNeuralNetwork, self).get_layer_descr(layer_idx) + child_str
    else:
      if isinstance(self.num_units_in_each_layer[layer_idx], (int, float)) and \
        np.isfinite(self.num_units_in_each_layer[layer_idx]):
        num_units_descr = ', ' + str(int(self.num_units_in_each_layer[layer_idx]))
      elif is_a_pooling_layer_label(self.layer_labels[layer_idx]):
        num_units_descr = ', 1'
      else:
        num_units_descr = ''
      if self.strides[layer_idx] == 2:
        stride_str = ' /2'
      else:
        stride_str = ''
      return '%d: %s%s%s\n(%d)'%(layer_idx, self.layer_labels[layer_idx], stride_str,
                                 num_units_descr, self.layer_masses[layer_idx])

  @classmethod
  def _get_child_layer_groups(cls):
    """ Returns layer groups for a CNN. """
    return ['conv', 'pool', 'fc']

  @classmethod
  def _get_layer_group_for_layer_label(cls, layer_label):
    """ Returns layer group for layer. """
    if (layer_label.startswith('conv') and layer_label != 'conv') or \
       (layer_label.startswith('res') and layer_label != 'res'):
      return 'conv'
    elif layer_label.endswith('pool') and layer_label != 'pool':
      return 'pool'
    elif layer_label in ['fc', 'softmax']:
      return layer_label
    else:
      return None


class MultiLayerPerceptron(NeuralNetwork):
  """ Implements a multi-layer perceptron. """
  #pylint: disable=attribute-defined-outside-init

  def __init__(self, class_or_reg, layer_labels, conn_mat, num_units_in_each_layer,
               all_layer_label_classes=None, layer_label_similarities=None):
    """ Constructor. """
    self.mandatory_child_attributes = []
    self.class_or_reg = class_or_reg
    if class_or_reg.lower().startswith('reg'):
      nn_class = 'mlp-reg'
    elif class_or_reg.lower().startswith('class'):
      nn_class = 'mlp-class'
    else:
      raise ValueError('class_or_reg should be either class or reg. Given %s.'%(
                       class_or_reg))
    super(MultiLayerPerceptron, self).__init__(nn_class, layer_labels, conn_mat,
      num_units_in_each_layer, all_layer_label_classes,
      layer_label_similarities)

  def _child_check_if_valid_network(self):
    """ Runs any checks relevant for a MLP. """
    # Test if all layers before the output are linear/softmax.
    last_layer_label = 'linear' if self.nn_class == 'mlp-reg' else 'softmax'
    return _check_if_layers_before_op_are(self.conn_mat, self.get_op_layer_idx(),
                                          self.layer_labels, last_layer_label)

  def _child_attrs_topological_sort(self, top_order):
    """ Topologically sort child attributes. """
    pass

  def _child_compute_layer_masses(self):
    """ Computes the layer masses and the number of channels at each layer. """
    self.num_channels_in_to_each_layer, self.num_channels_out_of_each_layer = \
      compute_num_channels_at_each_layer(self)
    self.layer_masses = compute_layer_masses(self.num_channels_in_to_each_layer,
      self.num_channels_out_of_each_layer, self.layer_labels)

  @classmethod
  def _get_child_layer_groups(cls):
    """ Returns layer groups for an MLP. """
    return ['rectifier', 'sigmoid']

  @classmethod
  def _get_layer_group_for_layer_label(cls, layer_label):
    """ Returns layer group for layer. """
    if layer_label in MLP_RECTIFIERS:
      return 'rectifier'
    elif layer_label in MLP_SIGMOIDS:
      return 'sigmoid'
    else:
      return None


# Some common functions ================================================================
def _get_common_layer_labels():
  """ Returns layer labels common for all types of neural networks. """
  return ['ip', 'op']

def get_cnn_layer_labels(max_conv_size=7):
  """ Gets the layer labels for a CNN. """
  conv_layer_labels = ['fc', 'max-pool', 'avg-pool', 'softmax']
  for conv_size in range(3, max_conv_size+1, 2):
    conv_layer_labels.append('conv%d'%(conv_size))
    conv_layer_labels.append('res%d'%(conv_size))
  return _get_common_layer_labels() + conv_layer_labels

def get_mlp_layer_labels(class_or_reg, list_of_activations=None):
  """ Gets the layer labels for a Multi-layer perceptron. """
  list_of_activations = MLP_RECTIFIERS + MLP_SIGMOIDS \
    if list_of_activations is None else list_of_activations
  if class_or_reg == 'class' and 'softmax' not in list_of_activations:
    list_of_activations.append('softmax')
  elif class_or_reg == 'reg' and 'linear' not in list_of_activations:
    list_of_activations.append('linear')
  else:
    raise ValueError('class_or_reg should be either class or reg. Given %s.'%(
                     class_or_reg))
  return _get_common_layer_labels() + list_of_activations




"""
  NUTS Sampler -- Adapted from pgmpy library: https://github.com/pgmpy/pgmpy
  -- kvysyara@andrew.cmu.edu
"""

# pylint: disable=invalid-name
# pylint: disable=no-self-use
# pylint: disable=arguments-differ
# pylint: disable=too-many-locals
# pylint: disable=unbalanced-tuple-unpacking

from __future__ import division

import numpy as np

# Local imports
from .base import LeapFrog
from .hmc import HamiltonianMCDA
from .check_functions import _check_1d_array_object

class NoUTurnSampler(HamiltonianMCDA):
  """
  Class for performing sampling in Continuous model
  using No U Turn Sampler (a variant of Hamiltonian Monte Carlo)

  Parameters:
  -----------
  model: Model from which sampling has to be done

  simulate_dynamics: Class to propose future states of position
      and momentum in time by simulating HamiltonianDynamics

  References
  ----------
  Matthew D. Hoffman, Andrew Gelman, The No-U-Turn Sampler: Adaptively
  Setting Path Lengths in Hamiltonian Monte Carlo. Journal of
  Machine Learning Research 15 (2014) 1351-1381
  """

  def __init__(self, model, simulate_dynamics=LeapFrog):

    super(NoUTurnSampler, self).__init__(model=model, simulate_dynamics=simulate_dynamics)

  def _initalize_tree(self, position, momentum, slice_var, stepsize):
    """
    Initalizes root node of the tree, i.e depth = 0
    """

    position_bar, momentum_bar, _ = self.simulate_dynamics(self.model, position,\
                                    momentum, stepsize).get_proposed_values()
    logp_bar = self.model.logp(position_bar)

    hamiltonian = logp_bar - 0.5 * np.dot(momentum_bar, momentum_bar)

    candidate_set_size = slice_var <= np.exp(hamiltonian)
    if slice_var > 0:
      accept_set_bool = hamiltonian > (np.log(slice_var) - 1000) # delta_max = 1000
    else:
      accept_set_bool = slice_var < np.exp(hamiltonian + 1000) # delta_max = 1000

    return position_bar, momentum_bar, candidate_set_size, accept_set_bool

  def _update_acceptance_criteria(self, position_forward, position_backward,\
                                  momentum_forward, momentum_backward, accept_set_bool,\
                                  candidate_set_size, candidate_set_size2):
    """
    Checks the acceptance criteria and updates it
    """

    criteria1 = np.dot((position_forward - position_backward), momentum_backward) >= 0

    criteria2 = np.dot((position_forward - position_backward), momentum_forward) >= 0

    accept_set_bool = accept_set_bool and criteria1 and criteria2
    candidate_set_size += candidate_set_size2

    return accept_set_bool, candidate_set_size

  def _build_tree(self, position, momentum, slice_var, direction, depth, stepsize):
    """
    Recursively builds a tree for proposing new position and momentum
    """

    # Parameter names in algorithm (here -> representation in algorithm)
    # position->theta, momentum->r, slice_var->u, direction->v, depth->j,
    # stepsize->epsilon, candidate_set_size->n, accept_set_bool->s
    if depth == 0:
      # Take single leapfrog step in the given direction (direction * stepsize)
      position_bar, momentum_bar, candidate_set_size, accept_set_bool =\
        self._initalize_tree(position, momentum, slice_var, direction * stepsize)

      return (position_bar, momentum_bar, position_bar, momentum_bar, position_bar,\
              candidate_set_size, accept_set_bool)

    else:
      # Build left and right subtrees
      (position_backward, momentum_backward, position_forward, momentum_forward,\
       position_bar, candidate_set_size, accept_set_bool) = self._build_tree(position,\
                                 momentum, slice_var, direction, depth - 1, stepsize)
      if accept_set_bool == 1:
        if direction == -1:
          # Build tree in backward direction
          (position_backward, momentum_backward, _, _, position_bar2,\
           candidate_set_size2, accept_set_bool2) = self._build_tree(position_backward,\
                          momentum_backward, slice_var, direction, depth - 1, stepsize)
        else:
          # Build tree in forward direction
          (_, _, position_forward, momentum_forward, position_bar2, candidate_set_size2,\
           accept_set_bool2) = self._build_tree(position_forward, momentum_forward,\
                              slice_var, direction, depth - 1, stepsize)

        if candidate_set_size2 + candidate_set_size:
          if np.random.rand() < \
              min(1, candidate_set_size2/(candidate_set_size2 + candidate_set_size)):
            position_bar = position_bar2
        else:
          position_bar = position_bar2

        accept_set_bool, candidate_set_size =\
          self._update_acceptance_criteria(position_forward, position_backward,\
                          momentum_forward, momentum_backward, accept_set_bool2,\
                          candidate_set_size, candidate_set_size2)

      return (position_backward, momentum_backward, position_forward, momentum_forward,
              position_bar, candidate_set_size, accept_set_bool)

  def _sample(self, position, stepsize):
    """
    Returns a sample using a single iteration of NUTS
    """

    # Re-sampling momentum
    momentum = np.random.normal(0, 1, len(position))

    # Initializations
    depth = 0
    position_backward, position_forward = position, position
    momentum_backward, momentum_forward = momentum, momentum
    candidate_set_size = accept_set_bool = 1
    log_pdf = self.model.logp(position)

    # Resample slice variable `u`
    slice_var = np.random.uniform(0, np.exp(log_pdf - 0.5 * np.dot(momentum, momentum)))

    while accept_set_bool == 1:
      direction = np.random.choice([-1, 1], p=[0.5, 0.5])
      if direction == -1:
        # Build a tree in backward direction
        (position_backward, momentum_backward, _, _, position_bar, candidate_set_size2,\
         accept_set_bool2) = self._build_tree(position_backward, momentum_backward,\
                                              slice_var, direction, depth, stepsize)
      else:
        # Build tree in forward direction
        (_, _, position_forward, momentum_forward, position_bar, candidate_set_size2,\
         accept_set_bool2) = self._build_tree(position_forward, momentum_forward,\
                                              slice_var, direction, depth, stepsize)
      if accept_set_bool2 == 1:
        if np.random.rand() < min(1, candidate_set_size2/candidate_set_size):
          position = position_bar.copy()

      accept_set_bool, candidate_set_size =\
              self._update_acceptance_criteria(position_forward, position_backward,\
              momentum_forward, momentum_backward, accept_set_bool2, candidate_set_size,\
              candidate_set_size2)
      depth += 1

    return position

  def sample(self, initial_pos, num_samples, burn=150, stepsize=None):
    """
    Method to return samples using No U Turn Sampler

    Parameters
    ----------
    initial_pos: A 1d array like object
        Vector representing values of parameter position, the starting
        state in markov chain.

    num_samples: int
        Number of samples to be generated

    burn: int, defaults to 150
        Number of samples to be ignored in the beginning

    stepsize: float , defaults to None
        The stepsize for proposing new values of position and momentum in
        simulate_dynamics. If None, then will be choosen suitably
    """

    if not hasattr(initial_pos, '__len__'):
      initial_pos = np.array([initial_pos])
    initial_pos = _check_1d_array_object(initial_pos, 'initial_pos')

    if stepsize is None:
      stepsize = self._find_reasonable_stepsize(initial_pos)

    samples = np.zeros([num_samples, len(initial_pos)], dtype=float)
    position_m = initial_pos

    for i in range(1, num_samples+burn+1):
      # Genrating sample
      position_m = self._sample(position_m, stepsize)
      # ignore first few samples
      if i >= burn:
        samples[i-burn-1] = position_m

    return samples


class NoUTurnSamplerDA(NoUTurnSampler):
  """
  Class for performing sampling in Continuous model
  using No U Turn sampler with dual averaging for
  adaptation of parameter stepsize.

  Parameters:
  -----------
  model: Model from which sampling has to be done

  simulate_dynamics: Class to propose future states of position
      and momentum in time by simulating HamiltonianDynamics

  delta: float (in between 0 and 1), defaults to 0.65
      The target HMC acceptance probability

  References
  ----------
  Matthew D. Hoffman, Andrew Gelman, The No-U-Turn Sampler: Adaptively
  Setting Path Lengths in Hamiltonian Monte Carlo. Journal of
  Machine Learning Research 15 (2014) 1351-1381
  """

  def __init__(self, model, simulate_dynamics=LeapFrog, delta=0.65):

    if not isinstance(delta, float) or delta > 1.0 or delta < 0.0:
      raise ValueError("delta should be a floating value in between 0 and 1")

    self.delta = delta
    super(NoUTurnSamplerDA, self).__init__(model=model,\
                                  simulate_dynamics=simulate_dynamics)

  def _build_tree(self, position, momentum, slice_var, direction, depth, stepsize,\
                  position0, momentum0):
    """
    Recursively builds a tree for proposing new position and momentum
    """
    if depth == 0:

      position_bar, momentum_bar, candidate_set_size, accept_set_bool =\
        self._initalize_tree(position, momentum, slice_var, direction * stepsize)

      #alpha =\
      #      min(1, self._acceptance_prob(position, position_bar, momentum, momentum_bar))
      acceptance_prob = self._acceptance_prob(position, position_bar, momentum,
                                              momentum_bar)
      alpha = 1
      if acceptance_prob < 0:
        alpha = np.exp(acceptance_prob)

      return (position_bar, momentum_bar, position_bar, momentum_bar, position_bar,\
              candidate_set_size, accept_set_bool, alpha, 1)

    else:
      (position_backward, momentum_backward, position_forward, momentum_forward,\
       position_bar, candidate_set_size, accept_set_bool, alpha, n_alpha) =\
       self._build_tree(position, momentum, slice_var, direction, depth - 1, stepsize,\
                                                                  position0, momentum0)

      if accept_set_bool == 1:
        if direction == -1:
          # Build tree in backward direction
          (position_backward, momentum_backward, _, _, position_bar2,
           candidate_set_size2, accept_set_bool2, alpha2, n_alpha2) =\
           self._build_tree(position_backward, momentum_backward, slice_var, direction,\
                                              depth - 1, stepsize, position0, momentum0)
        else:
          # Build tree in forward direction
          (_, _, position_forward, momentum_forward, position_bar2,
           candidate_set_size2, accept_set_bool2, alpha2, n_alpha2) =\
           self._build_tree(position_forward, momentum_forward, slice_var, direction,\
                                            depth - 1, stepsize, position0, momentum0)

        if (candidate_set_size2 + candidate_set_size) != 0:
          if np.random.rand() <\
                min(1, candidate_set_size2/(candidate_set_size2 + candidate_set_size)):
            position_bar = position_bar2
        else:
          position_bar = position_bar2

        alpha += alpha2
        n_alpha += n_alpha2
        accept_set_bool, candidate_set_size =\
        self._update_acceptance_criteria(position_forward, position_backward,\
        momentum_forward, momentum_backward, accept_set_bool2, candidate_set_size,\
                           candidate_set_size2)

      return (position_backward, momentum_backward, position_forward, momentum_forward,\
              position_bar, candidate_set_size, accept_set_bool, alpha, n_alpha)

  def _sample(self, position, stepsize):
    """
    Returns a sample using a single iteration of NUTS with dual averaging
    """

    # Re-sampling momentum
    momentum = np.random.normal(0, 1, len(position))

    # Initializations
    depth = 0
    position_backward, position_forward = position, position
    momentum_backward, momentum_forward = momentum, momentum
    candidate_set_size = accept_set_bool = 1
    position_m_1 = position
    log_pdf = self.model.logp(position)

    # Resample slice variable `u`
    slice_var = np.random.uniform(0, np.exp(log_pdf - 0.5 * np.dot(momentum, momentum)))

    while accept_set_bool == 1:
      direction = np.random.choice([-1, 1], p=[0.5, 0.5])
      if direction == -1:
        # Build a tree in backward direction
        (position_backward, momentum_backward, _, _, position_bar, candidate_set_size2,\
         accept_set_bool2, alpha, n_alpha) =\
         self._build_tree(position_backward, momentum_backward, slice_var, direction,\
                          depth, stepsize, position_m_1, momentum)
      else:
        # Build tree in forward direction
        (_, _, position_forward, momentum_forward, position_bar, candidate_set_size2,\
         accept_set_bool2, alpha, n_alpha) =\
         self._build_tree(position_forward, momentum_forward, slice_var, direction,\
                          depth, stepsize, position_m_1, momentum)

      # Sample position at random from C
      if accept_set_bool2 == 1:
        if np.random.rand() < min(1, candidate_set_size2/candidate_set_size):
          position = position_bar

      accept_set_bool, candidate_set_size =\
      self._update_acceptance_criteria(position_forward, position_backward,\
      momentum_forward, momentum_backward, accept_set_bool2, candidate_set_size,\
                                           candidate_set_size2)

      depth += 1

    return position, alpha, n_alpha

  def sample(self, initial_pos, num_samples, burn=150, num_adapt=10, stepsize=None):
    """
    Returns samples using No U Turn Sampler with dual averaging

    Parameters
    ----------
    initial_pos: A 1d array like object
        Vector representing values of parameter position, the starting
        state in markov chain.

    num_adapt: int
        The number of interations to run the adaptation of stepsize

    num_samples: int
        Number of samples to be generated

    burn: int, defaults to 150
        Number of samples to be ignored in the beginning

    stepsize: float, defaults to None
        The stepsize for proposing new values of position and momentum in
        simulate_dynamics. If None, then will be choosen suitably
    """

    if not hasattr(initial_pos, '__len__'):
      initial_pos = np.array([initial_pos])
    initial_pos = _check_1d_array_object(initial_pos, 'initial_pos')

    if stepsize is None:
      stepsize = self._find_reasonable_stepsize(initial_pos)

    if num_adapt <= 1:
      return NoUTurnSampler(self.model,\
                  self.simulate_dynamics).sample(initial_pos, num_samples, stepsize)

    mu = np.log(10.0 * stepsize)
    stepsize_bar = 1.0
    h_bar = 0.0

    samples = np.zeros([num_samples, len(initial_pos)], dtype=float)
    position_m = initial_pos

    for i in range(1, num_samples+burn+1):

      position_m, alpha, n_alpha = self._sample(position_m, stepsize)
      #ignore first few samples
      if i >= burn:
        samples[i-burn-1] = position_m

      if i <= num_adapt:
        stepsize, stepsize_bar, h_bar = self._adapt_params(stepsize, stepsize_bar, h_bar,\
                                                           mu, i, alpha, n_alpha)
      else:
        stepsize = stepsize_bar

    return samples



"""
  Unit tests for NUTS sampling.
  -- kvysyara@andrew.cmu.edu
"""
from __future__ import absolute_import

# pylint: disable=no-self-use
import unittest

import numpy as np

# Local imports
from .continuous import Normal, Beta, Exponential
from ..utils.base_test_class import BaseTestClass, execute_tests


@unittest.skip
class NutsTestCase(BaseTestClass):
  """ Unit tests for distributions in continuous.py """

  def setUp(self):
    """ Sets up unit tests. """
    self.size = 100000
    self.threshold = 0.1

  def _check_sample_sizes(self, samples):
    """ Compares the sample sizes with the size parameter"""
    assert self.size == len(samples)

  def _compute_mean(self, samples):
    """ Computes Mean """
    return np.mean(samples)

  def _compute_variance(self, samples):
    """ Computes Variance """
    return np.var(samples)

  def test_nuts_sampling_normal(self):
    """ Tests NUTS sampling from Normal distribution """
    self.report('Test NUTS sampling of Normal Distribution.')
    mean = 11
    variance = 3
    dist = Normal(mean, variance)
    samples = dist.draw_samples('nuts', self.size, np.array([0.1]), 10)
    mean_r = self._compute_mean(samples)
    var_r = self._compute_variance(samples)
    self._check_sample_sizes(samples)
    assert abs(mean - mean_r) <= self.threshold
    assert abs(variance - var_r) <= self.threshold
    self.report('%s :: test result: mean=%0.3f, variance=%0.3f'\
                 %(str(dist), mean_r, var_r), 'test_result')

  def test_nuts_sampling_beta(self):
    """ Tests nuts sampling from Beta Distribution """
    self.report('Test nuts sampling of Beta Distribution.')
    alpha = 1
    beta = 2
    dist = Beta(alpha, beta)
    mean = dist.get_mean()
    var = dist.get_variance()
    samples = dist.draw_samples('nuts', self.size, np.array([0.1]), 10)
    mean_r = self._compute_mean(samples)
    var_r = self._compute_variance(samples)
    self._check_sample_sizes(samples)
    assert (samples >= 0).all() and (samples <= 1).all()
    assert abs(mean - mean_r) <= self.threshold
    assert abs(var - var_r) <= self.threshold
    self.report('%s, mean=%0.3f, variance=%0.3f :: test result: mean=%0.3f, '
                'variance=%0.3f'%(str(dist), mean, var, mean_r, var_r), 'test_result')

  def test_nuts_sampling_exponential(self):
    """ Tests nuts sampling from Exponential distribution """
    self.report('Test nuts sampling of Exponential Distribution.')
    lam = 2
    dist = Exponential(lam)
    mean = dist.get_mean()
    var = dist.get_variance()
    samples = dist.draw_samples('nuts', self.size, np.array([0.1]), 10)
    mean_r = self._compute_mean(samples)
    var_r = self._compute_variance(samples)
    self._check_sample_sizes(samples)
    assert (samples >= 0).all()
    assert abs(mean - mean_r) <= self.threshold
    assert abs(var - var_r) <= self.threshold
    self.report('%s, mean=%0.3f, variance=%0.3f :: test result: mean=%0.3f, '
                'variance=%0.3f'%(str(dist), mean, var, mean_r, var_r), 'test_result')

if __name__ == '__main__':
  execute_tests()




"""
  Harness for GPs on cartesian product domains.
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=invalid-name
# pylint: disable=abstract-method
# pylint: disable=no-member

from __future__ import print_function
from argparse import Namespace
import numpy as np
# Local imports
from ..exd import domains
from ..exd.cp_domain_utils import load_cp_domain_from_config_file
from . import gp_core, mf_gp
from .euclidean_gp import get_euclidean_integral_gp_kernel_with_scale, \
                            prep_euclidean_integral_kernel_hyperparams
from .kernel import CartesianProductKernel, HammingKernel
from ..utils.general_utils import get_idxs_from_list_of_lists
from ..utils.option_handler import get_option_specs, load_options
from ..utils.reporters import get_reporter

# Part 1: Parameters and Arguments
# ================================

# Domain kernel parameters
_DFLT_DOMAIN_EUC_KERNEL_TYPE = 'matern'
_DFLT_DOMAIN_INT_KERNEL_TYPE = 'matern'
_DFLT_DOMAIN_DISCRETE_NUMERIC_KERNEL_TYPE = 'matern'
_DFLT_DOMAIN_DISCRETE_KERNEL_TYPE = 'hamming'
_DFLT_DOMAIN_NN_KERNEL_TYPE = 'otmann'
_DFLT_DOMAIN_MATERN_NU = 2.5
# Multi-fidelity parameters
_DFLT_FIDEL_EUC_KERNEL_TYPE = 'se'
_DFLT_FIDEL_INT_KERNEL_TYPE = 'se'
_DFLT_FIDEL_DISCRETE_NUMERIC_KERNEL_TYPE = 'se'
_DFLT_FIDEL_DISCRETE_KERNEL_TYPE = 'hamming'
_DFLT_FIDEL_MATERN_NU = 2.5

# Basic parameters
basic_cart_product_gp_args = [
  # For Euclidean domains ------------------------------------------------------------
  # These parameters will be shared by both Euclidean domains and Discrete Euclidean
  # domains.
  get_option_specs('dom_euc_kernel_type', False, 'default',
    'Kernel type for euclidean domains. '),
  get_option_specs('dom_euc_use_same_bandwidth', False, False,
    ('If true, will use same bandwidth on all dimensions. Applies only '
     'when kernel_type is se or matern. Default=False.')),
  get_option_specs('dom_euc_matern_nu', False, 'default', \
    'Specify nu value for matern kernel. If negative, will fit.'),
  get_option_specs('dom_euc_poly_order', False, 1,
    'Order of the polynomial kernle to be used for Euclidean domains. ' + \
    'Default is 1 (linear kernel).'),
  # Additive models for Euclidean spaces
  get_option_specs('dom_euc_use_additive_gp', False, False,
    'Whether or not to use an additive GP. '),
  get_option_specs('dom_euc_add_max_group_size', False, 6,
    'The maximum number of groups in the additive grouping. '),
  get_option_specs('dom_euc_add_grouping_criterion', False, 'randomised_ml',
    'Specify the grouping algorithm, should be one of {randomised_ml}'),
  get_option_specs('dom_euc_num_groups_per_group_size', False, -1,
    'The number of groups to try per group size.'),
  get_option_specs('dom_euc_add_group_size_criterion', False, 'sampled',
    'Specify how to pick the group size, should be one of {max,sampled}.'),
  # ESP for Euclidean spaces
  get_option_specs('dom_euc_esp_order', False, -1,
    'Order of the esp kernel. '),
  get_option_specs('dom_euc_esp_kernel_type', False, 'se',
    'Specify type of kernel. This depends on the application.'),
  get_option_specs('dom_euc_esp_matern_nu', False, 'default', \
    ('Specify the nu value for matern kernel. If negative, will fit.')),
  # For Integral domains --------------------------------------------------------------
  get_option_specs('dom_int_kernel_type', False, 'default',
    'Kernel type for integral domains. '),
  get_option_specs('dom_int_use_same_bandwidth', False, False,
    ('If true, will use same bandwidth on all dimensions. Applies only '
     'when kernel_type is se or matern. Default=False.')),
  get_option_specs('dom_int_matern_nu', False, 'default', \
    'Specify nu value for matern kernel. If negative, will fit.'),
  get_option_specs('dom_int_poly_order', False, 1,
    'Order of the polynomial kernle to be used for Integral domains. ' + \
    'Default is 1 (linear kernel).'),
  # Additive models for Integral spaces
  get_option_specs('dom_int_use_additive_gp', False, False,
    'Whether or not to use an additive GP. '),
  get_option_specs('dom_int_add_max_group_size', False, 6,
    'The maximum number of groups in the additive grouping. '),
  get_option_specs('dom_int_add_grouping_criterion', False, 'randomised_ml',
    'Specify the grouping algorithm, should be one of {randomised_ml}'),
  get_option_specs('dom_int_num_groups_per_group_size', False, -1,
    'The number of groups to try per group size.'),
  get_option_specs('dom_int_add_group_size_criterion', False, 'sampled',
    'Specify how to pick the group size, should be one of {max,sampled}.'),
  # ESP for Integral spaces
  get_option_specs('dom_int_esp_order', False, -1,
    'Order of the esp kernel. '),
  get_option_specs('dom_int_esp_kernel_type', False, 'se',
    'Specify type of kernel. This depends on the application.'),
  get_option_specs('dom_int_esp_matern_nu', False, 'default', \
    ('Specify the nu value for matern kernel. If negative, will fit.')),
  # For Discrete Numeric domains --------------------------------------------------------
  get_option_specs('dom_disc_num_kernel_type', False, 'default',
    'Kernel type for discrete numeric domains. '),
  get_option_specs('dom_disc_num_use_same_bandwidth', False, False,
    ('If true, will use same bandwidth on all dimensions. Applies only '
     'when kernel_type is se or matern. Default=False.')),
  get_option_specs('dom_disc_num_matern_nu', False, 'default', \
    'Specify nu value for matern kernel. If negative, will fit.'),
  get_option_specs('dom_disc_num_poly_order', False, 1,
    'Order of the polynomial kernle to be used for Integral domains. ' + \
    'Default is 1 (linear kernel).'),
  # ESP for Discrete Numeric spaces
  get_option_specs('dom_disc_num_esp_order', False, -1,
    'Order of the esp kernel. '),
  get_option_specs('dom_disc_num_esp_kernel_type', False, 'se',
    'Specify type of kernel. This depends on the application.'),
  get_option_specs('dom_disc_num_esp_matern_nu', False, 'default', \
    ('Specify the nu value for matern kernel. If negative, will fit.')),
  # For Discrete domains -----------------------------------------------------------------
  get_option_specs('dom_disc_kernel_type', False, 'default',
    'Kernel type for discrete domains.'),
  get_option_specs('dom_disc_hamming_use_same_weight', False, False,
    'If true, use same weight for all dimensions of the hamming kernel.'),
  # For NN domains -----------------------------------------------------------------------
  get_option_specs('dom_nn_kernel_type', False, 'default',
    'Kernel type for NN Domains.'),
  # Arguments specific to OTMANN ---------------------------------------------------------
  get_option_specs('otmann_dist_type', False, 'lp-emd',
    'The type of distance. Should be lp, emd or lp-emd.'),
  get_option_specs('otmann_kernel_type', False, 'lpemd_sum',
    'The Otmann kernel type. Should be one of lp, emd, lpemd_sum, or lpemd_prod.'),
  get_option_specs('otmann_choose_mislabel_struct_coeffs', False, 'use_given',
    ('How to choose the mislabel and struct coefficients. Should be one of ' +
     'tune_coeffs or use_given. In the latter case, otmann_mislabel_coeffs and ' +
     'otmann_struct_coeffs should be non-empty.')),
  get_option_specs('otmann_mislabel_coeffs', False, '1.0-1.0-1.0-1.0',
    'The mislabel coefficients specified as a string. If -1, it means we will tune.'),
  get_option_specs('otmann_struct_coeffs', False, '0.1-0.25-0.61-1.5',
    'The struct coefficients specified as a string. If -1, it means we will tune.'),
  get_option_specs('otmann_lp_power', False, 1,
    'The powers to use in the LP distance for the kernel.'),
  get_option_specs('otmann_emd_power', False, 2,
    'The powers to use in the EMD distance for the kernel.'),
  get_option_specs('otmann_non_assignment_penalty', False, 1.0,
    'The non-assignment penalty for the OTMANN distance.'),
  ]

cartesian_product_gp_args = gp_core.mandatory_gp_args + basic_cart_product_gp_args

basic_mf_cart_product_gp_args = [ \
  # Fidelity kernel Euclidean -----------------------------------------------------------
  get_option_specs('fidel_euc_kernel_type', False, 'se', \
    ('Type of kernel for the Euclidean part of the fidelity space. Should be se, ' +
     'matern, poly or expdecay')),
  get_option_specs('fidel_euc_matern_nu', False, 2.5, \
    ('Specify the nu value for the matern kernel. If negative, will fit.')),
  get_option_specs('fidel_euc_use_same_bandwidth', False, False, \
    ('If true, will use same bandwidth on all Euclidean fidelity dimensions. Applies ' +
     'only when fidel_kernel_type is se or matern. Default=False.')),
  # Fidelity kernel Integral -----------------------------------------------------------
  get_option_specs('fidel_int_kernel_type', False, 'se', \
    'Type of kernel for the fidelity space. Should be se, matern, poly or expdecay'),
  get_option_specs('fidel_int_matern_nu', False, 2.5, \
    ('Specify the nu value for the matern kernel. If negative, will fit.')),
  get_option_specs('fidel_int_use_same_bandwidth', False, False, \
    ('If true, will use same bandwidth on all integral fidelity dimensions. Applies ' +
     'only when fidel_kernel_type is se or matern. Default=False.')),
  # Fidelity kernel Discrete Numeric ---------------------------------------------------
  get_option_specs('fidel_disc_num_kernel_type', False, 'se', \
    'Type of kernel for the fidelity space. Should be se, matern, poly or expdecay'),
  get_option_specs('fidel_disc_num_matern_nu', False, 2.5, \
    ('Specify the nu value for the matern kernel. If negative, will fit.')),
  get_option_specs('fidel_disc_num_use_same_bandwidth', False, False, \
    ('If true, will use same bandwidth on all integral fidelity dimensions. Applies ' +
     'only when fidel_kernel_type is se or matern. Default=False.')),
  # Fidelity Kernel Discrete ----------------------------------------------------------
  get_option_specs('fidel_disc_kernel_type', False, 'default',
    'Kernel type for discrete domains.'),
  get_option_specs('fidel_disc_hamming_use_same_weight', False, False,
    'If true, use same weight for all dimensions of the hamming kernel.'),
  ]

cartesian_product_mf_gp_args = cartesian_product_gp_args + basic_mf_cart_product_gp_args


def get_default_kernel_type(domain_type):
  """ Returns default kernel type for the domain. """
  if domain_type == 'euclidean':
    return _DFLT_DOMAIN_EUC_KERNEL_TYPE
  elif domain_type == 'discrete_euclidean':
    return _DFLT_DOMAIN_EUC_KERNEL_TYPE
  elif domain_type == 'integral':
    return _DFLT_DOMAIN_INT_KERNEL_TYPE
  elif domain_type == 'prod_discrete':
    return _DFLT_DOMAIN_DISCRETE_KERNEL_TYPE
  elif domain_type == 'prod_discrete_numeric':
    return _DFLT_DOMAIN_DISCRETE_NUMERIC_KERNEL_TYPE
  elif domain_type == 'neural_network':
    return _DFLT_DOMAIN_NN_KERNEL_TYPE
  else:
    raise ValueError('Unknown domain_type: %s.'%(domain_type))


# Part 2: CPGP and CPGPFitter
# ====================================================================================
class CPGP(gp_core.GP):
  """ GP class for Cartesian Product GPs. """
  # pylint: disable=attribute-defined-outside-init

  def __init__(self, X, Y, kernel, mean_func, noise_var, domain_lists_of_dists=None,
               build_posterior=True, reporter=None,
               handle_non_psd_kernels='project_first'):
    """
      Constructor
      X, Y: data
      kernel: a kernel object
    """
    if domain_lists_of_dists is None:
      domain_lists_of_dists = [None] * kernel.num_kernels
    self.domain_lists_of_dists = domain_lists_of_dists
    super(CPGP, self).__init__(X, Y, kernel, mean_func, noise_var, build_posterior,
                               reporter, handle_non_psd_kernels)

  def set_domain_lists_of_dists(self, domain_lists_of_dists):
    """ Sets lists of distances. """
    self.domain_lists_of_dists = domain_lists_of_dists

  def _child_str(self):
    """ String representation of the GP. """
    if len(self.X) > 0:
      mean_str = 'mu=%0.4f, '%(self.mean_func([self.X[0]])[0])
    else:
      mean_str = ''
    return mean_str + str(self.kernel)

  def _get_training_kernel_matrix(self):
    """ Returns the training kernel matrix. """
    n = len(self.X)
    ret = self.kernel.hyperparams['scale'] * np.ones((n, n))
    for idx, kern in enumerate(self.kernel.kernel_list):
      if self.domain_lists_of_dists[idx] is not None:
        ret *= kern.evaluate_from_dists(self.domain_lists_of_dists[idx])
      else:
        curr_X = get_idxs_from_list_of_lists(self.X, idx)
        ret *= kern(curr_X, curr_X)
    return ret


class CPMFGP(mf_gp.MFGP):
  """ Multi-fidelity GP Class for Cartesian Product GPs. """

  def __init__(self, ZZ, XX, YY, mf_kernel, mean_func, noise_var,
               kernel_scale=None, fidel_space_kernel=None, domain_kernel=None,
               fidel_space_lists_of_dists=None, domain_lists_of_dists=None,
               build_posterior=True, reporter=None,
               handle_non_psd_kernels='project_first'):
    """ Constructor. """
    # pylint: disable=too-many-arguments
    if mf_kernel is None:
      mf_kernel = CartesianProductKernel(kernel_scale,
                                         [fidel_space_kernel, domain_kernel])
    self.kernel_scale = kernel_scale
    self.fidel_space_kernel = fidel_space_kernel
    self.domain_kernel = domain_kernel
    # Now check the lists_of_dists
    if fidel_space_lists_of_dists is None:
      fidel_space_lists_of_dists = [None] * mf_kernel[0].num_kernels
    if domain_lists_of_dists is None:
      domain_lists_of_dists = [None] * mf_kernel[0].num_kernels
    self.fidel_space_lists_of_dists = fidel_space_lists_of_dists
    self.domain_lists_of_dists = domain_lists_of_dists
    # Finally call the super constructor.
    super(CPMFGP, self).__init__(ZZ, XX, YY, mf_kernel, mean_func, noise_var,
                                 build_posterior=build_posterior,
                                 reporter=reporter,
                                 handle_non_psd_kernels=handle_non_psd_kernels)

  def _child_str(self):
    """ String representation of the GP. """
    if len(self.X) > 0:
      mean_str = 'mu=%0.4f, '%(self.mean_func([self.X[0]])[0])
    else:
      mean_str = ''
    return mean_str + str(self.kernel)

  def set_fidel_space_lists_of_dists(self, fidel_space_lists_of_dists):
    """ Sets the fidel_space lists of dists. """
    self.fidel_space_lists_of_dists = fidel_space_lists_of_dists

  def set_domain_lists_of_dists(self, domain_lists_of_dists):
    """ Sets the domain lists of dists. """
    self.domain_lists_of_dists = domain_lists_of_dists

  @classmethod
  def _get_train_kernel_matrix_from_data_and_lists_of_dists(cls, data, kern,
                                                            lists_of_dists):
    """ Returns the training kernel matrix given the data, kernel, and lists_of_dists. """
    n = len(data)
    ret = kern.hyperparams['scale'] * np.ones((n, n))
    for idx, kern in enumerate(kern.kernel_list):
      if lists_of_dists[idx] is not None:
        ret *= kern.evaluate_from_dists(lists_of_dists[idx])
      else:
        curr_data = get_idxs_from_list_of_lists(data, idx)
        ret *= kern(curr_data, curr_data)
    return ret

  def _get_training_kernel_matrix(self):
    """ Returns the training kernel matrix. """
    if self.fidel_space_kernel is None or self.domain_kernel is None:
      return self.kernel(self.X, self.X)
    else:
      KF = self._get_train_kernel_matrix_from_data_and_lists_of_dists(self.ZZ,
             self.fidel_space_kernel, self.fidel_space_lists_of_dists)
      KD = self._get_train_kernel_matrix_from_data_and_lists_of_dists(self.XX,
             self.domain_kernel, self.domain_lists_of_dists)
      return self.kernel_scale * KF * KD


# Part 3: GP Fitters
# ================================

class CPGPFitter(gp_core.GPFitter):
  """ Fits a cartesian product GP by tuning the hyper-parameters. """

  def __init__(self, X, Y, domain, domain_kernel_ordering=None,
               domain_lists_of_dists=None, domain_dist_computers=None,
               options=None, reporter=None):
    """ Constructor. """
    # pylint: disable=no-member
    # Set domain
    if isinstance(domain, str):
      domain, config_orderings = load_cp_domain_from_config_file(domain)
      if domain_kernel_ordering is None:
        domain_kernel_ordering = config_orderings.kernel_ordering
    # Check before assignment
    if not isinstance(domain, domains.Domain):
      raise ValueError('domain should be instance of domain.Domain or a file name' +
                       ' to a configuration json/protobuf file describing the domain.')
    if domain_kernel_ordering is None:
      raise ValueError('If domain is not a config file name, domain_kernel_ordering' +
                       ' cannot be None.')
    self.domain_num_domains = domain.num_domains
    self.domain = domain
    self.domain_kernel_ordering = domain_kernel_ordering
    # Now create a struct to store kernel dependent parameters
    if domain_lists_of_dists is None:
      domain_lists_of_dists = [None] * self.domain_num_domains
    if domain_dist_computers is None:
      domain_dist_computers = [None] * self.domain_num_domains
    self.domain_lists_of_dists = domain_lists_of_dists
    self.domain_dist_computers = domain_dist_computers
    self.domain_kernel_params_for_each_domain = [Namespace() for _ in
                                          range(self.domain_num_domains)]
    for idx in range(self.domain_num_domains):
      self.domain_kernel_params_for_each_domain[idx].list_of_dists = \
        domain_lists_of_dists[idx]
      self.domain_kernel_params_for_each_domain[idx].dist_computer = \
        domain_dist_computers[idx]
    # Create reporter and options
    reporter = get_reporter(reporter)
    if options is None:
      options = load_options(cartesian_product_gp_args, 'CPGP', reporter=reporter)
    # Call super constructor
    super(CPGPFitter, self).__init__(X, Y, options, reporter)

  def _child_set_up(self):
    """ Set up parameters for the CPGPFitter. """
    # Set up for the kernel scale
    self.param_order.append(['kernel_scale', 'cts'])
    self.kernel_scale_log_bounds = [np.log(0.03 * self.Y_var), np.log(30 * self.Y_var)]
    self.cts_hp_bounds.append(self.kernel_scale_log_bounds)
    _set_up_hyperparams_for_domain(self, self.X, self.domain, 'dom',
                                   self.domain_kernel_ordering,
                                   self.domain_kernel_params_for_each_domain,
                                   self.domain_dist_computers,
                                   self.domain_lists_of_dists)

  # Methods to build the GP ------------------------------------------------------
  def _child_build_gp(self, mean_func, noise_var, gp_cts_hps, gp_dscr_hps,
                      other_gp_params=None, *args, **kwargs):
    """ Builds a Domain product GP. """
    kernel_scale = np.exp(gp_cts_hps[0])
    gp_cts_hps = gp_cts_hps[1:]
    cp_kernel, gp_cts_hps, gp_dscr_hps = _build_kernel_for_domain(self.domain, 'dom',
      kernel_scale, gp_cts_hps, gp_dscr_hps, other_gp_params, self.options,
      self.domain_kernel_ordering, self.domain_kernel_params_for_each_domain)
    ret_gp = CPGP(self.X, self.Y, cp_kernel, mean_func, noise_var,
                  domain_lists_of_dists=self.domain_lists_of_dists, *args, **kwargs)
    return ret_gp, gp_cts_hps, gp_dscr_hps


class CPMFGPFitter(mf_gp.MFGPFitter):
  """ A Fitter for multi-fidelity GPs in Cartesian Product spaces. """

  def __init__(self, ZZ, XX, YY, config=None,
               fidel_space=None, domain=None,
               fidel_space_kernel_ordering=None, domain_kernel_ordering=None,
               fidel_space_lists_of_dists=None, domain_lists_of_dists=None,
               fidel_space_dist_computers=None, domain_dist_computers=None,
               options=None, reporter=None):
    """ Constructor. """
    # pylint: disable=too-many-arguments
    # Load options
    reporter = get_reporter(reporter)
    if options is None:
      options = load_options(cartesian_product_mf_gp_args)
    # Read from config
    if config is not None:
      if isinstance(config, str):
        from ..exd.exd_utils import load_config_file
        config = load_config_file(config)
      self.config = config
      self.fidel_space = config.fidel_space
      self.domain = config.domain
      self.fidel_space_kernel_ordering = config.fidel_space_orderings.kernel_ordering
      self.domain_kernel_ordering = config.domain_orderings.kernel_ordering
      self.fidel_space_num_domains = self.fidel_space.num_domains
    elif fidel_space is not None and domain is not None and \
      fidel_space_kernel_ordering is not None and domain_kernel_ordering is not None:
      self.config = None
      self.fidel_space = fidel_space
      self.domain = domain
      self.fidel_space_kernel_ordering = fidel_space_kernel_ordering
      self.domain_kernel_ordering = domain_kernel_ordering
    else:
      raise ValueError('Provide either config or parameters fidel_space, domain' +
                       ' fidel_space_kernel_ordering, and domain_kernel_ordering.')
    self.fidel_space_num_domains = self.fidel_space.num_domains
    self.domain_num_domains = self.domain.num_domains
    # Create a structure to store kernel parameters
    self.fidel_space_kernel_params_for_each_domain = \
      [Namespace() for _ in range(self.fidel_space_num_domains)]
    self.domain_kernel_params_for_each_domain = \
      [Namespace() for _ in range(self.domain_num_domains)]
    # Store lists of dists & dist_computers
    if fidel_space_lists_of_dists is None:
      fidel_space_lists_of_dists = [None] * self.fidel_space_num_domains
    if domain_lists_of_dists is None:
      domain_lists_of_dists = [None] * self.domain_num_domains
    if fidel_space_dist_computers is None:
      fidel_space_dist_computers = [None] * self.fidel_space_num_domains
    if domain_dist_computers is None:
      domain_dist_computers = [None] * self.domain_num_domains
    self.fidel_space_lists_of_dists = fidel_space_lists_of_dists
    self.domain_lists_of_dists = domain_lists_of_dists
    self.fidel_space_dist_computers = fidel_space_dist_computers
    self.domain_dist_computers = domain_dist_computers
    for idx in range(self.fidel_space_num_domains):
      self.fidel_space_kernel_params_for_each_domain[idx].list_of_dists = \
        fidel_space_lists_of_dists[idx]
      self.fidel_space_kernel_params_for_each_domain[idx].dist_computer = \
        fidel_space_dist_computers[idx]
    for idx in range(self.domain_num_domains):
      self.domain_kernel_params_for_each_domain[idx].list_of_dists = \
        domain_lists_of_dists[idx]
      self.domain_kernel_params_for_each_domain[idx].dist_computer = \
        domain_dist_computers[idx]
    # Call super constructor
    super(CPMFGPFitter, self).__init__(ZZ, XX, YY, options, reporter)

  # Child set up
  def _child_set_up(self):
    """ Sets parameters for the GPFitter. """
    self.param_order.append(['kernel_scale', 'cts'])
    self.kernel_scale_log_bounds = [np.log(0.03 * self.Y_var), np.log(30 * self.Y_var)]
    self.cts_hp_bounds.append(self.kernel_scale_log_bounds)
    _set_up_hyperparams_for_domain(self, self.ZZ, self.fidel_space, 'fidel',
                                   self.fidel_space_kernel_ordering,
                                   self.fidel_space_kernel_params_for_each_domain,
                                   self.fidel_space_dist_computers,
                                   self.fidel_space_lists_of_dists)
    _set_up_hyperparams_for_domain(self, self.XX, self.domain, 'dom',
                                   self.domain_kernel_ordering,
                                   self.domain_kernel_params_for_each_domain,
                                   self.domain_dist_computers,
                                   self.domain_lists_of_dists)

  def _child_build_gp(self, mean_func, noise_var, gp_cts_hps, gp_dscr_hps,
                      other_gp_params=None, *args, **kwargs):
    """ Builds a Cartesian product Multi-fidelity GP. """
    kernel_scale = np.exp(gp_cts_hps[0])
    gp_cts_hps = gp_cts_hps[1:]
    # Create kernels
    fidel_space_kernel, gp_cts_hps, gp_dscr_hps = _build_kernel_for_domain(
      self.fidel_space, 'fidel', 1.0, gp_cts_hps, gp_dscr_hps, other_gp_params,
      self.options, self.fidel_space_kernel_ordering,
      self.fidel_space_kernel_params_for_each_domain)
    domain_kernel, gp_cts_hps, gp_dscr_hps = _build_kernel_for_domain(self.domain, 'dom',
      1.0, gp_cts_hps, gp_dscr_hps, other_gp_params, self.options,
      self.domain_kernel_ordering, self.domain_kernel_params_for_each_domain)
    # build the GP
    ret_gp = CPMFGP(self.ZZ, self.XX, self.YY, None, mean_func, noise_var,
                    kernel_scale, fidel_space_kernel, domain_kernel,
                    self.fidel_space_lists_of_dists, self.domain_lists_of_dists,
                    *args, **kwargs)
    return ret_gp, gp_cts_hps, gp_dscr_hps


# Part 4: Some utilities we will use above
# ========================================

# 4.1 Utilities to set up GP Hyperparams -------------------------------------------------
def _set_up_hyperparams_for_domain(fitter, X_data, gp_domain, dom_prefix,
  kernel_ordering, kernel_params_for_each_domain, dist_computers, lists_of_dists):
  """ This modifies the fitter object. """
  # pylint: disable=too-many-branches
  for dom_idx, dom, kernel_type in \
    zip(range(gp_domain.num_domains), gp_domain.list_of_domains, kernel_ordering):
    dom_type = dom.get_type()
    dom_identifier = '%s-%d-%s'%(dom_prefix, dom_idx, dom_type)
    # Kernel type
    if kernel_type == '' or kernel_type is None:
      # If none is specified, use the one given in options
      kernel_type = _get_kernel_type_from_options(dom_type, dom_prefix,
                                                  fitter.options)
    if kernel_type == 'default':
      kernel_type = get_default_kernel_type(dom.get_type())
    # Iterate through each individual domain and add it to the hyper parameters
    curr_dom_Xs = get_idxs_from_list_of_lists(X_data, dom_idx)
    # Some conditional options
    if dom_type in ['euclidean', 'integral', 'prod_discrete_numeric',
                    'discrete_euclidean']:
      use_same_bw, matern_nu, esp_kernel_type, esp_matern_nu = \
        _get_euc_int_options(dom_type, dom_prefix, fitter.options)
      # Now set things up depending on the kernel type
      if kernel_type == 'se':
        fitter.cts_hp_bounds, fitter.param_order = _set_up_dim_bandwidths(
          dom_identifier, curr_dom_Xs, use_same_bw, dom.get_dim(),
          fitter.cts_hp_bounds, fitter.param_order)
      elif kernel_type == 'matern':
        if isinstance(matern_nu, float) and matern_nu < 0:
          fitter.dscr_hp_vals.append([0.5, 1.5, 2.5])
          fitter.param_order.append(['%s-%s'%(dom_identifier, 'matern_nu'), 'dscr'])
        fitter.cts_hp_bounds, fitter.param_order = _set_up_dim_bandwidths(
          dom_identifier, curr_dom_Xs, use_same_bw, dom.get_dim(),
          fitter.cts_hp_bounds, fitter.param_order)
      elif kernel_type == 'poly':
        raise NotImplementedError('Not implemented polynomial kernels yet.')
      elif kernel_type == 'expdecay':
        # Offset
        scale_range = fitter.Y_var / np.sqrt(fitter.num_tr_data + 0.0001)
        fitter.cts_hp_bounds.append(
          [np.log(0.1 * scale_range), np.log(10 * scale_range)])
        fitter.param_order.append(['%s-expdecay_log_offset'%(dom_identifier), 'cts'])
        # Power
        fitter.cts_hp_bounds.extend([[np.log(1e-1), np.log(50)]] * dom.get_dim())
        identifiers = [['%s-expdecay_log_power-%d'%(dom_identifier, i), 'cts']
                       for i in range(dom.get_dim())]
        fitter.param_order.extend(identifiers)
      elif kernel_type == 'esp':
        fitter.cts_hp_bounds, fitter.param_order = _set_up_dim_bandwidths(
          dom_identifier, curr_dom_Xs, use_same_bw, dom.get_dim(),
          fitter.cts_hp_bounds, fitter.param_order)
        # Matern nu
        if esp_kernel_type == 'matern' and esp_matern_nu < 0:
          fitter.dscr_hp_vals.append([0.5, 1.5, 2.5])
          fitter.param_order.append(['%s-%s'%(dom_identifier, 'esp_matern_nu'), 'dscr'])
        # ESP order
        esp_order_vals = range(1, dom.get_dim() // 2)
        fitter.param_order.append(['%s-%s'%(dom_identifier, 'esp_order'), 'dscr'])
        fitter.dscr_hp_vals.append(esp_order_vals)
      else:
        raise ValueError('Unknown kernel type "%s" for "%s" spaces.'%(kernel_type,
                         dom.get_type()))
    elif dom.get_type() in ['prod_discrete']:
      if kernel_type == 'hamming':
        disc_hamming_use_same_weight, = _get_disc_options(dom_prefix, fitter.options)
        fitter.cts_hp_bounds, fitter.param_order = _set_up_hamming_weights(dom_identifier,
          disc_hamming_use_same_weight, dom.get_dim(),
          fitter.cts_hp_bounds, fitter.param_order)
      else:
        raise ValueError('Unknown kernel type "%s" for "%s" spaces.'%(kernel_type,
                         dom.get_type()))
    elif dom.get_type() in ['neural_network']:
      if kernel_type == 'otmann':
        fitter.cts_hp_bounds, fitter.param_order = \
          _set_up_nn_domain_otmann(dom_idx, dom, dom_identifier, curr_dom_Xs,
            fitter.options, fitter.cts_hp_bounds, fitter.param_order,
            kernel_params_for_each_domain, dist_computers, lists_of_dists)
      else:
        raise ValueError('Unknown kernel type "%s" for "%s" spaces.'%(kernel_type,
                         dom.get_type()))


def _get_euc_int_options(dom_type, dom_prefix, options):
  """ Returns fieds from options depending on dom_type. """
  dom_type_code_dic = {'euclidean': 'euc',
                       'integral': 'int',
                       'prod_discrete_numeric': 'disc_num',
                       'discrete_euclidean': 'euc',
                      }
  def _extract_from_options(dom_type_str, property_str):
    """ Extracts the property from the dom_type. """
    attr_name_in_options = dom_prefix + '_' + dom_type_code_dic[dom_type_str] + \
                           '_' + property_str
    if hasattr(options, attr_name_in_options):
      return getattr(options, attr_name_in_options)
    else:
      return None
  # Now extract each property that we need
  use_same_bw = _extract_from_options(dom_type, 'use_same_bandwidth')
  matern_nu = _extract_from_options(dom_type, 'matern_nu')
  esp_kernel_type = _extract_from_options(dom_type, 'esp_kernel_type')
  esp_matern_nu = _extract_from_options(dom_type, 'esp_matern_nu')
  return use_same_bw, matern_nu, esp_kernel_type, esp_matern_nu


def _get_disc_options(dom_prefix, options):
  """ Returns the options, depending on dom_type. """
  def _extract_from_options(property_str):
    """ Extracts property from options. """
    attr_name_in_options = dom_prefix + '_' + property_str
    if hasattr(options, attr_name_in_options):
      return getattr(options, attr_name_in_options)
    else:
      return None
  disc_hamming_use_same_weight = _extract_from_options('disc_hamming_use_same_weight')
  return (disc_hamming_use_same_weight, )


def _get_kernel_type_from_options(dom_type, dom_prefix, options):
  """ Returns kernel type from options. """
  dom_type_descr_dict = {'euclidean': 'euc',
                         'discrete_euclidean': 'euc',
                         'integral': 'int',
                         'prod_discrete_numeric': 'disc_num',
                         'prod_discrete': 'disc',
                         'neural_network': 'nn',
                        }
  if dom_type not in dom_type_descr_dict.keys():
    raise ValueError('Unknown domain type %s.'%(dom_type))
  attr_name = '%s_%s_kernel_type'%(dom_prefix, dom_type_descr_dict[dom_type])
  return getattr(options, attr_name)


def _set_up_dim_bandwidths(dom_identifier, curr_dom_Xs, use_same_bandwidth, dim,
                           cts_hp_bounds, param_order):
  """ Set up for dim bandwidths. """
  hp_identifier = '%s-%s'%(dom_identifier, 'dom_bandwidths')
  if len(curr_dom_Xs) > 0:
    # If there is data
    assert len(curr_dom_Xs[0]) == dim # temporary: check this condition
    curr_dom_Xs = np.array(curr_dom_Xs)
    curr_dom_mean_diffs = curr_dom_Xs - curr_dom_Xs.mean(axis=0)
    curr_dom_X_std_norms = [np.linalg.norm(curr_dom_mean_diffs[:, idx]) + 1e-4
                            for idx in range(dim)]
  else:
    curr_dom_X_std_norms = [1.0] * dim
  if use_same_bandwidth:
    bandwidth_log_bounds = [np.log(0.01 * curr_dom_X_std_norms.min()),
                            np.log(100 * curr_dom_X_std_norms.max())]
    cts_hp_bounds.append(bandwidth_log_bounds)
    param_order.append([hp_identifier + '-same_bandwidth', 'cts'])
  else:
    bandwidth_log_bounds = [[np.log(0.01 * elem), np.log(100 * elem)] \
                            for elem in curr_dom_X_std_norms]
    cts_hp_bounds.extend(bandwidth_log_bounds)
    param_order.extend([[hp_identifier + '-%d'%(idx), 'cts']
                         for idx in range(len(bandwidth_log_bounds))])
  return cts_hp_bounds, param_order


def _set_up_hamming_weights(dom_identifier, use_same_weights, dim,
                            cts_hp_bounds, param_order):
  """ Set up for dim_weights in hamming kernel. """
  if use_same_weights or dim == 1:
    return cts_hp_bounds, param_order
  elif dim == 2:
    cts_hp_bounds.append([0, 1])
    param_order.append([dom_identifier + '-hamming_wt-2D', 'cts'])
  elif dim > 2:
    cts_hp_bounds.extend([[0, 1]] * dim)
    param_order.extend([['%s-hamming_wts-%d'%(dom_identifier, idx), 'cts']
                             for idx in range(dim)])
  else:
    raise ValueError('dim=%0.4f should be a positive integer.'%(dim))
  return cts_hp_bounds, param_order


def _set_up_nn_domain_otmann(dom_idx, dom, dom_identifier, curr_dom_Xs, options,
  cts_hp_bounds, param_order, kernel_params_for_each_domain, dist_computers,
  lists_of_dists):
  """ Set up for NN Domain. """
  _process_otmann_kernel_params(dom_idx, options, kernel_params_for_each_domain,
                                dist_computers)
  # Check if we need to pre-compute distances -----------------------------------------
  if options.otmann_choose_mislabel_struct_coeffs == 'use_given' and \
    kernel_params_for_each_domain[dom_idx].list_of_dists is None:
    # Then compute the distances
    if dist_computers[dom_idx] is None:
      from ..nn.otmann import get_otmann_distance_computer_from_args
      curr_dist_computer = get_otmann_distance_computer_from_args(
        dom.nn_type, options.otmann_non_assignment_penalty,
        kernel_params_for_each_domain[dom_idx].otmann_mislabel_coeffs,
        kernel_params_for_each_domain[dom_idx].otmann_struct_coeffs,
        kernel_params_for_each_domain[dom_idx].otmann_dist_type)
    else:
      curr_dist_computer = dist_computers[dom_idx]
    # Add the current distance computer to dist_computers
    kernel_params_for_each_domain[dom_idx].otmann_distance_computer = \
      curr_dist_computer
    dist_computers[dom_idx] = curr_dist_computer
    # Now compute lists of dists
    curr_list_of_dists = curr_dist_computer(curr_dom_Xs, curr_dom_Xs)
    kernel_params_for_each_domain[dom_idx].list_of_dists = \
      curr_list_of_dists
    lists_of_dists[dom_idx] = curr_list_of_dists
  # Create bounds here ----------------------------------------------------------------
  # 1. Relative weights for LP/EMD terms
  if kernel_params_for_each_domain[dom_idx].otmann_kernel_type == 'lpemd_sum':
    cts_hp_bounds.append([0, 1]) # controls relative weighting of each term
    param_order.append([dom_identifier + '-lp_emd_tradeoff', 'cts'])
  # 2. Betas
  lp_beta_log_bounds = [[np.log(1e-9), np.log(1e-3)]] * \
    kernel_params_for_each_domain[dom_idx].otmann_num_mislabel_struct_coeffs
  emd_beta_log_bounds = [[np.log(1e-1), np.log(1e2)]] * \
    kernel_params_for_each_domain[dom_idx].otmann_num_mislabel_struct_coeffs
  if kernel_params_for_each_domain[dom_idx].otmann_dist_type == 'lp':
    all_beta_bounds = lp_beta_log_bounds
  elif kernel_params_for_each_domain[dom_idx].otmann_dist_type == 'emd':
    all_beta_bounds = emd_beta_log_bounds
  elif kernel_params_for_each_domain[dom_idx].otmann_dist_type == 'lp-emd':
    all_beta_bounds = [j for i in zip(lp_beta_log_bounds, emd_beta_log_bounds)
                       for j in i]
  cts_hp_bounds.extend(all_beta_bounds)
  param_order.extend([[dom_identifier + '-beta-%d'%(idx), 'cts']
                       for idx in range(len(all_beta_bounds))])
  # 3 & 4. mislabel/struct coeffs
  if kernel_params_for_each_domain[dom_idx].otmann_to_tune_mislabel_struct_coeffs:
    cts_hp_bounds.append([0.001, 2.0]) # mislabel coefficient (not in log space)
    param_order.append([dom_identifier + '-mislabel_coeff', 'cts'])
    cts_hp_bounds.append([0.001, 2.0]) # structural coefficient (not in log space)
    param_order.append([dom_identifier + '-struct_coeff', 'cts'])
  return cts_hp_bounds, param_order


def _process_otmann_kernel_params(dom_idx, options, kernel_params_for_each_domain,
                                  dist_computers):
  """ Preprocesses the structural and mislabel coefficients for domain in dom_idx. """
  # First check if dist_type and kernel_type are consistent ----------------------------
  otmann_dist_type = options.otmann_dist_type
  otmann_kernel_type = options.otmann_kernel_type
  otmann_choose_mislabel_struct_coeffs = \
    options.otmann_choose_mislabel_struct_coeffs
  otmann_mislabel_coeffs = options.otmann_mislabel_coeffs
  otmann_struct_coeffs = options.otmann_struct_coeffs
  if otmann_dist_type in ['lp', 'emd'] and otmann_kernel_type != otmann_dist_type:
    raise ValueError('If dist_type is %s, then kernel_type should be %s.'%(
                     otmann_dist_type, otmann_dist_type))
  elif otmann_dist_type == 'lp-emd' and \
       otmann_kernel_type not in ['lpemd_prod', 'lpemd_sum']:
    raise ValueError('If otmann_dist_type is lp-emd, then otmann_kernel_type should ' +
                     'be lpemd_sum or lpemd_prod.')
  # Check if the mislabel/struct coeffs are give as appropriate
  if otmann_choose_mislabel_struct_coeffs == 'use_given' and \
    (otmann_mislabel_coeffs == '' or otmann_struct_coeffs == ''):
    raise ValueError(('If choose_mislabel_struct_coeffs is use_given, then ' +
      'mislabel_coeffs and struct_coeffs cannot be empty. Given mislabel_coeffs=%s,' +
      ' struct_coeffs=%s')%(otmann_mislabel_coeffs, otmann_struct_coeffs))
  # Now process the mislabel and struct coeffs -----------------------------------------
  if otmann_choose_mislabel_struct_coeffs == 'tune_coeffs':
    # If they are integers or floats
    otmann_num_mislabel_struct_coeffs = 1
    otmann_to_tune_mislabel_struct_coeffs = True
  elif (isinstance(otmann_mislabel_coeffs, list) and
        isinstance(otmann_struct_coeffs, list)) or \
       (isinstance(otmann_mislabel_coeffs, str) and
        isinstance(otmann_struct_coeffs, str)):
    # If they are lists or strings
    if isinstance(otmann_mislabel_coeffs, str):
      # If they are strings, obtain the lists
      otmann_mislabel_coeffs = [float(x) for x in otmann_mislabel_coeffs.split('-')]
      otmann_struct_coeffs = [float(x) for x in otmann_struct_coeffs.split('-')]
    else:
      otmann_mislabel_coeffs = otmann_mislabel_coeffs
      otmann_struct_coeffs = otmann_struct_coeffs
    # Other parameters
    otmann_num_mislabel_struct_coeffs = len(otmann_mislabel_coeffs)
    otmann_to_tune_mislabel_struct_coeffs = False
    # Check if the lengths are the same
    if len(otmann_mislabel_coeffs) != len(otmann_struct_coeffs):
      raise ValueError('Length of mislabel and structural coefficients must be same.' +
        'Given: mislabel: %s (%d), struct:%s (%d).'%(
        str(otmann_mislabel_coeffs), len(otmann_mislabel_coeffs),
        str(otmann_struct_coeffs), len(otmann_struct_coeffs)))
  else:
    raise ValueError('Bad format for mislabel and structural coefficients.' +
      'Given: mislabel: %s, struct:%s.'%(
      str(otmann_mislabel_coeffs), str(otmann_struct_coeffs)))
  # Now store them in kernel_params_for_each_domain ------------------------------------
  kernel_params_for_each_domain[dom_idx].otmann_mislabel_coeffs = \
    otmann_mislabel_coeffs
  kernel_params_for_each_domain[dom_idx].otmann_struct_coeffs = \
    otmann_struct_coeffs
  kernel_params_for_each_domain[dom_idx].otmann_num_mislabel_struct_coeffs = \
    otmann_num_mislabel_struct_coeffs
  kernel_params_for_each_domain[dom_idx].otmann_to_tune_mislabel_struct_coeffs = \
    otmann_to_tune_mislabel_struct_coeffs
  kernel_params_for_each_domain[dom_idx].otmann_choose_mislabel_struct_coeffs = \
    otmann_choose_mislabel_struct_coeffs
  kernel_params_for_each_domain[dom_idx].otmann_dist_type = otmann_dist_type
  kernel_params_for_each_domain[dom_idx].otmann_kernel_type = otmann_kernel_type
  # store other parameters from options
  kernel_params_for_each_domain[dom_idx].otmann_lp_power = \
    options.otmann_lp_power
  kernel_params_for_each_domain[dom_idx].otmann_emd_power = \
    options.otmann_emd_power
  kernel_params_for_each_domain[dom_idx].otmann_non_assignment_penalty = \
    options.otmann_non_assignment_penalty
  # Add the distance computer
  kernel_params_for_each_domain[dom_idx].otmann_distance_computer = \
    dist_computers[dom_idx]


# 4.2 Utilities to build the GP --------------------------------------------------------
def _build_kernel_for_domain(domain, dom_prefix, kernel_scale, gp_cts_hps, gp_dscr_hps,
  other_gp_params, options, kernel_ordering, kernel_params_for_each_domain):
  """ Builds the kernel for the domain. """
  kernel_list = []
  # Iterate through each domain and build the corresponding kernel
  for dom_idx, dom, kernel_type in \
    zip(range(domain.num_domains), domain.list_of_domains,
        kernel_ordering):
    dom_type = dom.get_type().lower()
    if kernel_type == '' or kernel_type is None:
      # If none is specified, use the one given in options
      kernel_type = _get_kernel_type_from_options(dom_type, 'dom', options)
    if kernel_type == 'default':
      kernel_type = get_default_kernel_type(dom.get_type())
    if dom_type in ['euclidean', 'integral', 'prod_discrete_numeric',
                    'discrete_euclidean']:
      curr_kernel_hyperparams = _prep_kernel_hyperparams_for_euc_int_kernels(
                                  kernel_type, dom, dom_prefix, options)
      use_same_bw, _, esp_kernel_type, _ = \
        _get_euc_int_options(dom_type, 'dom', options)
      if hasattr(other_gp_params, 'add_gp_groupings') and \
        other_gp_params.add_gp_groupings is not None:
        add_gp_groupings = other_gp_params.add_gp_groupings[dom_idx]
      else:
        add_gp_groupings = None
      curr_kernel, gp_cts_hps, gp_dscr_hps = \
        get_euclidean_integral_gp_kernel_with_scale(kernel_type, 1.0, \
          curr_kernel_hyperparams, gp_cts_hps, gp_dscr_hps, use_same_bw,
          add_gp_groupings, esp_kernel_type)
    elif dom_type == 'prod_discrete':
      curr_kernel_hyperparams = _prep_kernel_hyperparams_for_discrete_kernels(
                                     kernel_type, dom, dom_prefix, options)
      curr_kernel, gp_cts_hps, gp_dscr_hps = \
        get_discrete_kernel(kernel_type, curr_kernel_hyperparams, gp_cts_hps,
                            gp_dscr_hps)
    elif dom_type == 'neural_network':
      curr_kernel_hyperparams = _prep_kernel_hyperparams_for_nn_kernels(
                                  kernel_type, dom,
                                  kernel_params_for_each_domain[dom_idx])
      curr_kernel, gp_cts_hps, gp_dscr_hps = \
        get_neural_network_kernel(kernel_type, curr_kernel_hyperparams, gp_cts_hps,
                                  gp_dscr_hps)
    else:
      raise NotImplementedError(('Not implemented _child_build_gp for dom_type ' +
                                 '%s yet.')%(dom_type))
    kernel_list.append(curr_kernel)
  return CartesianProductKernel(kernel_scale, kernel_list), gp_cts_hps, gp_dscr_hps


def _prep_kernel_hyperparams_for_euc_int_kernels(kernel_type, dom, dom_prefix, options):
  """ Prepares the kernel hyperparams. """
  # This function retrieves a particular value from the options
  def _get_option_val_from_args(dom_type_str, attr_name):
    """ Internal function to get the options. """
    attr_name_in_options = dom_prefix + '_' + dom_type_str + '_' + attr_name
    if hasattr(options, attr_name_in_options):
      return getattr(options, attr_name_in_options)
    else:
      return None
  # This function retrieves several relevant values in a namepace.
  def _get_options_for_domain(dom_type_str):
    """ Internal function to return options for the domain. """
    default_matern_nu = _DFLT_DOMAIN_MATERN_NU if dom_prefix == 'dom' else \
                        _DFLT_FIDEL_MATERN_NU
    # set them in namespace
    matern_nu_in_options = _get_option_val_from_args(dom_type_str, 'matern_nu')
    matern_nu = default_matern_nu if matern_nu_in_options == 'default' else \
                  matern_nu_in_options
    esp_matern_nu_in_options = _get_option_val_from_args(dom_type_str, 'esp_matern_nu')
    esp_matern_nu = default_matern_nu if esp_matern_nu_in_options == 'default' else \
                  esp_matern_nu_in_options
    euc_int_options = Namespace(matern_nu=matern_nu,
                        esp_matern_nu=esp_matern_nu,
                        poly_order=_get_option_val_from_args(dom_type_str, 'poly_order'),
                        esp_order=_get_option_val_from_args(dom_type_str, 'esp_order'),
                        )
    return euc_int_options

  # Call above functions with relevant arguments.
  if dom.get_type() in ['euclidean', 'discrete_euclidean']:
    euc_int_options = _get_options_for_domain('euc')
  elif dom.get_type() == 'integral':
    euc_int_options = _get_options_for_domain('int')
  elif dom.get_type() == 'prod_discrete_numeric':
    euc_int_options = _get_options_for_domain('disc_num')
  else:
    raise ValueError('kernel_type should be euclidean, integral or ' +
                     'prod_discrete_numeric in this function.')
  return prep_euclidean_integral_kernel_hyperparams(kernel_type, euc_int_options,
                                                    dom.get_dim())

def _prep_kernel_hyperparams_for_discrete_kernels(kernel_type, dom, dom_prefix, options):
  """ Prepares the kernel hyperparams. """
  hamming_wt_attr_name = dom_prefix + '_' + 'disc_hamming_use_same_weight'
  ret = {'disc_hamming_use_same_weight': getattr(options, hamming_wt_attr_name),
         'dim': dom.get_dim(),
         'kernel_type': kernel_type,
        }
  return ret

def _prep_kernel_hyperparams_for_nn_kernels(kernel_type, dom, kernel_params_for_dom):
  """ Prepares the kernel hyper-parameters. """
  ret = vars(kernel_params_for_dom)
  ret['nn_type'] = dom.nn_type
  ret['kernel_type'] = kernel_type
  return ret

def get_discrete_kernel(kernel_type, kernel_hyperparams, gp_cts_hps, gp_dscr_hps):
  """ Returns the kernel for discrete hyper-parameters. """
  dim = kernel_hyperparams['dim']
  if kernel_type == 'hamming':
    if dim == 1 or kernel_hyperparams['disc_hamming_use_same_weight']:
      dim_wts = np.ones((dim,))/float(dim)
    elif dim == 2:
      dim_1_wt = gp_cts_hps[0]
      dim_wts = np.array([dim_1_wt, 1 - dim_1_wt])
      gp_cts_hps = gp_cts_hps[1:]
    else:
      dim_wts_unnorm = np.array(gp_cts_hps[0:dim])
      dim_wts_sum = dim_wts_unnorm.sum()
      if dim_wts_sum > 0:
        dim_wts = dim_wts_unnorm / dim_wts_sum
      else:
        dim_wts = np.ones((dim,))/float(dim)
      gp_cts_hps = gp_cts_hps[dim:]
    kern = HammingKernel(dim_wts)
  else:
    raise ValueError('Unknown kernel_type "%s" for discrete spaces.'%(kernel_type))
  return kern, gp_cts_hps, gp_dscr_hps


def get_neural_network_kernel(kernel_type, kernel_hyperparams, gp_cts_hps,
                              gp_dscr_hps):
  """ Returns the kernel for the Neural Network. """
  # pylint: disable=too-many-branches
  if kernel_type == 'otmann':
    # 1. Relative weights
    if kernel_hyperparams['otmann_kernel_type'] == 'lpemd_sum':
      lp_emd_tradeoff = gp_cts_hps[0]
      lp_emd_alphas = [lp_emd_tradeoff, 1 - lp_emd_tradeoff]
      gp_cts_hps = gp_cts_hps[1:]
    # 2. Betas
    if kernel_hyperparams['otmann_dist_type'] in ['lp', 'emd']:
      betas = np.exp(gp_cts_hps[:kernel_hyperparams['otmann_num_mislabel_struct_coeffs']])
      gp_cts_hps = gp_cts_hps[kernel_hyperparams['otmann_num_mislabel_struct_coeffs']:]
    elif kernel_hyperparams['otmann_dist_type'] == 'lp-emd':
      betas = np.exp(gp_cts_hps[:
                2 * kernel_hyperparams['otmann_num_mislabel_struct_coeffs']])
      gp_cts_hps = gp_cts_hps[
                     2 * kernel_hyperparams['otmann_num_mislabel_struct_coeffs']:]
    # 3 & 4. Mislabel/struct coeffs
    if kernel_hyperparams['otmann_to_tune_mislabel_struct_coeffs']:
      mislabel_coeffs = [gp_cts_hps[0]]
      struct_coeffs = [gp_cts_hps[1]]
      gp_cts_hps = gp_cts_hps[2:]
    else:
      mislabel_coeffs = kernel_hyperparams['otmann_mislabel_coeffs']
      struct_coeffs = kernel_hyperparams['otmann_struct_coeffs']
    # Determine the powers
    if kernel_hyperparams['otmann_dist_type'] == 'lp':
      powers = [kernel_hyperparams['otmann_lp_power']] * \
                kernel_hyperparams['otmann_num_mislabel_struct_coeffs']
    elif kernel_hyperparams['otmann_dist_type'] == 'emd':
      powers = [kernel_hyperparams['otmann_emd_power']] * \
                kernel_hyperparams['otmann_num_mislabel_struct_coeffs']
    elif kernel_hyperparams['otmann_dist_type'] == 'lp-emd':
      powers = [kernel_hyperparams['otmann_lp_power'],
                kernel_hyperparams['otmann_emd_power']] * \
                kernel_hyperparams['otmann_num_mislabel_struct_coeffs']
    else:
      raise ValueError('Unknown distance type: %s'%(
                        kernel_hyperparams['otmann_dist_type']))
    # Check if there is a distance computer
    if 'otmann_distance_computer' in kernel_hyperparams and \
      kernel_hyperparams['otmann_distance_computer'] is not None:
      tp_comp = kernel_hyperparams['otmann_distance_computer']
    else:
      from ..nn.otmann import get_otmann_distance_computer_from_args
      tp_comp = get_otmann_distance_computer_from_args(kernel_hyperparams['nn_type'],
                                 kernel_hyperparams['otmann_non_assignment_penalty'],
                                 mislabel_coeffs, struct_coeffs,
                                 kernel_hyperparams['otmann_dist_type'])
    # Now create the kernel
    from ..nn.otmann import DistProdNNKernel, DistSumNNKernel
    if kernel_hyperparams['otmann_kernel_type'] in ['lpemd_prod', 'lp', 'sum']:
      kern = DistProdNNKernel(tp_comp, betas, 1.0, powers)
    elif kernel_hyperparams['otmann_kernel_type'] in ['lpemd_sum']:
      kern = DistSumNNKernel(tp_comp, lp_emd_alphas, betas, powers)
    else:
      raise ValueError('Unknown otmann_kernel_type: %s.'%(
                        kernel_hyperparams['otmann_kernel_type']))
  else:
    raise ValueError('Unknown kernel_type %s.'%(kernel_type))
  # return
  return kern, gp_cts_hps, gp_dscr_hps




"""
  Harness to manage optimisation domains.
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=invalid-name
# pylint: disable=arguments-differ

import numpy as np
from numbers import Number
from scipy.spatial.distance import cdist

class Domain(object):
  """ Domain class. An abstract class which implements domains. """

  def get_type(self):
    """ Returns the type of the domain. """
    raise NotImplementedError('Implement in a child class.')

  def get_dim(self):
    """ Returns the dimension of the space. """
    raise NotImplementedError('Implement in a child class.')

  def is_a_member(self, point):
    """ Returns True if point is a member of this domain. """
    raise NotImplementedError('Implement in a child class.')

  @classmethod
  def members_are_equal(cls, point_1, point_2):
    """ Compares two members and returns True if they are the same. """
    return point_1 == point_2

  def compute_distance(self, point_1, point_2):
    """ Computes the distance between point_1 and point_2. """
    raise NotImplementedError('Implement in a child class.')

  def __str__(self):
    """ Returns a string representation. """
    raise NotImplementedError('Implement in a child class.')


# Universal Domain ----------
class UniversalDomain(Domain):
  """ A universal domian. Everything is a part of this.
      Used mostly in instances where the domain is not critical for lazy coding.
  """

  def get_type(self):
    """ Returns the type of the domain. """
    return 'universal'

  def get_dim(self):
    """ Return the dimensions. """
    return None

  def is_a_member(self, _):
    """ Returns true if point is in the domain. """
    return True

  @classmethod
  def compute_distance(cls, point_1, point_2):
    """ Computes the distance between point_1 and point_2. """
    raise ValueError('Distance not defined for Universal Domain.')

  def __str__(self):
    """ Returns a string representation. """
    return 'Universal Domain'


# Euclidean spaces ---------
class EuclideanDomain(Domain):
  """ Domain for Euclidean spaces. """

  def __init__(self, bounds):
    """ Constructor. """
    _check_if_valid_euc_int_bounds(bounds)
    self.bounds = np.array(bounds)
    self.diameter = np.linalg.norm(self.bounds[:, 1] - self.bounds[:, 0])
    self.dim = len(bounds)
    super(EuclideanDomain, self).__init__()

  def get_type(self):
    """ Returns the type of the domain. """
    return 'euclidean'

  def get_dim(self):
    """ Return the dimensions. """
    return self.dim

  def is_a_member(self, point):
    """ Returns true if point is in the domain. """
    return is_within_bounds(self.bounds, point)

  def members_are_equal(self, point_1, point_2):
    """ Compares two members and returns True if they are the same. """
    return self.compute_distance(point_1, point_2) < 1e-8 * self.diameter

  @classmethod
  def compute_distance(cls, point_1, point_2):
    """ Computes the distance between point_1 and point_2. """
    return np.linalg.norm(np.array(point_1) - np.array(point_2))

  def __str__(self):
    """ Returns a string representation. """
    return 'Euclidean: %s'%(_get_bounds_as_str(self.bounds))


# Integral spaces ------------
class IntegralDomain(Domain):
  """ Domain for vector valued integers. """

  def __init__(self, bounds):
    """ Constructor. """
    _check_if_valid_euc_int_bounds(bounds)
    self.bounds = np.array(bounds, dtype=np.int)
    self.diameter = np.linalg.norm(self.bounds[:, 1] - self.bounds[:, 0])
    self.dim = len(bounds)
    super(IntegralDomain, self).__init__()

  def get_type(self):
    """ Returns the type of the domain. """
    return 'integral'

  def get_dim(self):
    """ Return the dimensions. """
    return self.dim

  def is_a_member(self, point):
    """ Returns true if point is in the domain. """
    are_ints = [isinstance(x, (int, np.int, np.int64)) for x in point]
    return all(are_ints) and is_within_bounds(self.bounds, point)

  def members_are_equal(self, point_1, point_2):
    """ Compares two members and returns True if they are the same. """
    return self.compute_distance(point_1, point_2) < 1e-8 * self.diameter

  @classmethod
  def compute_distance(cls, point_1, point_2):
    """ Computes the distance between point_1 and point_2. """
    return np.linalg.norm(np.array(point_1) - np.array(point_2))

  def __str__(self):
    """ Returns a string representation. """
    return 'Integral: %s'%(_get_bounds_as_str(self.bounds))


# Discrete spaces -------------
class DiscreteDomain(Domain):
  """ A Domain for discrete objects. """

  def __init__(self, list_of_items):
    """ Constructor. """
    self.list_of_items = list_of_items
    self.size = len(list_of_items)
    super(DiscreteDomain, self).__init__()

  def get_type(self):
    """ Returns the type of the domain. """
    return 'discrete'

  def get_dim(self):
    """ Return the dimensions. """
    return 1

  def is_a_member(self, point):
    """ Returns true if point is in the domain. """
    return point in self.list_of_items

  @classmethod
  def _get_disc_domain_type(cls):
    """ Prefix for __str__. Can be overridden by a child class. """
    return "Disc"

  @classmethod
  def compute_distance(cls, point_1, point_2):
    """ Computes the distance between point_1 and point_2. """
    return float(point_1 != point_2)

  def __str__(self):
    """ Returns a string representation. """
    base_str = '%s(%d)'%(self._get_disc_domain_type(), self.size)
    if self.size < 4:
      return '%s: %s'%(base_str, self.list_of_items)
    return base_str


class DiscreteNumericDomain(DiscreteDomain):
  """ A domain for discrete objects all of which are numeric. """

  def __init__(self, list_of_items):
    """ Constructor. """
    if not all_items_are_numeric(list_of_items):
      raise ValueError('list_of_items must be a list of numbers.')
    super(DiscreteNumericDomain, self).__init__(list_of_items)

  def get_type(self):
    """ Returns the type of the domain. """
    return 'discrete_numeric'

  def _get_disc_domain_type(self):
    """ Prefix for __str__. Can be overridden by a child class. """
    return "DiscNum"

  @classmethod
  def compute_distance(cls, point_1, point_2):
    """ Computes the distance between point_1 and point_2. """
    return abs(point_1 - point_2)

  def is_a_member(self, point):
    """ Returns true if point is in the domain. """
    return discrete_numeric_element_is_in_list(point, self.list_of_items)


class DiscreteEuclideanDomain(DiscreteDomain):
  """ Domain for Discrete Euclidean spaces. """

  def __init__(self, list_of_items):
    """ Constructor. """
    list_of_items = np.array(list_of_items)
    self.dim = list_of_items.shape[1]
    self.size = len(list_of_items)
    self.diameter = np.sqrt(self.dim) * (list_of_items.max() - list_of_items.min())
    super(DiscreteEuclideanDomain, self).__init__(list_of_items)

  def get_type(self):
    """ Returns the type of the domain. """
    return 'discrete_euclidean'

  def _get_disc_domain_type(self):
    """ Prefix for __str__. Can be overridden by a child class. """
    return "DiscEuc"

  def get_dim(self):
    """ Return the dimensions. """
    return self.dim

  @classmethod
  def compute_distance(cls, point_1, point_2):
    """ Computes the distance between point_1 and point_2. """
    return np.linalg.norm(np.array(point_1) - np.array(point_2))

  def is_a_member(self, point):
    """ Returns true if point is in the domain. """
    # Naively find the nearest point in the domain
    return cdist([point], self.list_of_items).min() < 1e-8 * self.diameter

  def members_are_equal(self, point_1, point_2):
    """ Compares two members and returns True if they are the same. """
    return self.compute_distance(point_1, point_2) < 1e-8 * self.diameter


# A product of discrete spaces -----------------------------------------------------
class ProdDiscreteDomain(Domain):
  """ A product of discrete objects. """

  def __init__(self, list_of_list_of_items):
    """ Constructor. """
    self.list_of_list_of_items = list_of_list_of_items
    self.dim = len(list_of_list_of_items)
    self.size = np.prod([len(loi) for loi in list_of_list_of_items])

  def get_type(self):
    """ Returns the type of the domain. """
    return 'prod_discrete'

  def get_dim(self):
    """ Return the dimensions. """
    return self.dim

  def is_a_member(self, point):
    """ Returns true if point is in the domain. """
    if not hasattr(point, '__iter__') or len(point) != self.dim:
      return False
    ret = [elem in loi for elem, loi in zip(point, self.list_of_list_of_items)]
    return all(ret)

  def members_are_equal(self, point_1, point_2):
    """ Compares two members and returns True if they are the same. """
    elems_are_equal = [point_1[i] == point_2[i] for i in range(self.dim)]
    return all(elems_are_equal)

  @classmethod
  def _get_prod_disc_domain_type(cls):
    """ Prefix for __str__. Can be overridden by a child class. """
    return "ProdDisc"

  @classmethod
  def compute_distance(cls, point_1, point_2):
    """ Computes the distance between point_1 and point_2. """
    return float(sum([elem_1 != elem_2 for (elem_1, elem_2) in zip(point_1, point_2)]))

  def __str__(self):
    """ Returns a string representation. """
    return '%s(d=%d,size=%d)'%(self._get_prod_disc_domain_type(), self.dim, self.size)


class ProdDiscreteNumericDomain(ProdDiscreteDomain):
  """ A product of discrete numeric objects. """

  def __init__(self, list_of_list_of_items):
    """ Constructor. """
    if not all_lists_of_items_are_numeric(list_of_list_of_items):
      raise ValueError('list_of_list_of_items must of a list where each element is ' +
                       'a list of numeric objects.')
    super(ProdDiscreteNumericDomain, self).__init__(list_of_list_of_items)

  def get_type(self):
    """ Returns the type of the domain. """
    return 'prod_discrete_numeric'

  def is_a_member(self, point):
    """ Returns True if point is in the domain. """
    if not hasattr(point, '__iter__') or len(point) != self.dim:
      return False
    ret = [discrete_numeric_element_is_in_list(elem, loi)
           for elem, loi in zip(point, self.list_of_list_of_items)]
    return all(ret)

  @classmethod
  def compute_distance(cls, point_1, point_2):
    """ Computes the distance between point_1 and point_2. """
    return np.linalg.norm(np.array(point_1) - np.array(point_2))

  @classmethod
  def _get_prod_disc_domain_type(cls):
    """ Prefix for __str__. Can be overridden by a child class. """
    return "ProdDiscNum"


# Compound Domains ------------------------------------------
# Implementing a series of domains derived from the above

class CartesianProductDomain(Domain):
  """ The cartesian product of several domains. """

  def __init__(self, list_of_domains, domain_info=None):
    """ Constructor.
        list_of_domains is a list of domain objects.
        An element in this domain is represented by a list whose ith element
        belongs to list_of_domains[i].
    """
    self.list_of_domains = list_of_domains
    self.num_domains = len(list_of_domains)
    try:
      self.dim = sum([dom.get_dim() for dom in self.list_of_domains])
    except TypeError:
      self.dim = None
    # Domain info
    self.domain_info = domain_info
    self._has_constraints = False
    if self.domain_info is not None:
      from .cp_domain_utils import get_raw_point_from_processed_point
      self.raw_name_ordering = self.domain_info.config_orderings.raw_name_ordering
      self.get_raw_point = lambda x: get_raw_point_from_processed_point(x,
                             self, self.domain_info.config_orderings.index_ordering,
                             self.domain_info.config_orderings.dim_ordering)
      if hasattr(self.domain_info, 'config_file') and \
        self.domain_info.config_file is not None:
        import os
        self.config_file = self.domain_info.config_file
        self.config_file_dir = os.path.dirname(os.path.abspath(os.path.realpath(
                                               self.domain_info.config_file)))
      if hasattr(self.domain_info, 'constraints'):
        self._has_constraints = True
        self._constraint_eval_set_up()

  def _constraint_eval_set_up(self):
    """ Set up for evaluating constraints. """
    from importlib import import_module
    import sys
    from ..utils.general_utils import evaluate_strings_with_given_variables
    self.str_constraint_evaluator = evaluate_strings_with_given_variables
    self.domain_constraints = self.domain_info.constraints
    self.num_domain_constraints = len(self.domain_constraints)
    # Separate the constraints into different types
    self.eval_as_pyfile_idxs = [idx for idx in range(self.num_domain_constraints) if
                                isinstance(self.domain_constraints[idx][1], str) and
                                self.domain_constraints[idx][1].endswith('.py')]
    self.eval_as_str_idxs = [idx for idx in range(self.num_domain_constraints) if
                             isinstance(self.domain_constraints[idx][1], str) and
                             idx not in self.eval_as_pyfile_idxs]
    self.eval_as_pyfunc_idxs = [idx for idx in range(self.num_domain_constraints) if
                                hasattr(self.domain_constraints[idx][1], '__call__')]
    # Save constraints here
    self.pyfunc_constraints = [self.domain_constraints[idx][1] for idx
                              in self.eval_as_pyfunc_idxs]
    self.str_constraints = [self.domain_constraints[idx][1] for idx in
                            self.eval_as_str_idxs]
    # pyfile constraints
    pyfile_modules = [self.domain_constraints[idx][1] for idx
                      in self.eval_as_pyfile_idxs]
    self.pyfile_constraints = []
    sys.path.append(self.config_file_dir)
    for pfm_file_name in pyfile_modules:
      pfm = pfm_file_name.split('.')[0]
      constraint_source_module = import_module(pfm, self.config_file_dir)
      self.pyfile_constraints.append(constraint_source_module.constraint)
    sys.path.remove(self.config_file_dir)

  def get_type(self):
    """ Returns the type of the domain. """
    return 'cartesian_product'

  def has_constraints(self):
    """ Returns True if the domain has constraints. """
    return self._has_constraints

  def get_dim(self):
    """ Returns the dimension. """
    return self.dim

  def is_a_member(self, point):
    """ Returns true if the point is in the domain. """
    if not hasattr(point, '__iter__') or len(point) != self.num_domains:
      return False
    for dom_pt, dom in zip(point, self.list_of_domains):
      if not dom.is_a_member(dom_pt): # check if each element is in the respective domain.
        return False
    # Now check if the constraints are satisfied
    if not self.constraints_are_satisfied(point):
      return False
    return True

  def _evaluate_all_constraints(self, raw_point, name_to_pt_dict):
    """ Evaluates all constraints. """
    # Evaluate all constraints
    ret_str_all = self.str_constraint_evaluator(self.str_constraints, name_to_pt_dict)
    ret_pyfile_all = [elem(raw_point) for elem in self.pyfile_constraints]
    ret_pyfunc_all = [elem(raw_point) for elem in self.pyfunc_constraints]
    # Merge results
    ret_all = [None] * self.num_domain_constraints
    for str_idx, orig_idx in enumerate(self.eval_as_str_idxs):
      ret_all[orig_idx] = ret_str_all[str_idx]
    for pyfile_idx, orig_idx in enumerate(self.eval_as_pyfile_idxs):
      ret_all[orig_idx] = ret_pyfile_all[pyfile_idx]
    for pyfunc_idx, orig_idx in enumerate(self.eval_as_pyfunc_idxs):
      ret_all[orig_idx] = ret_pyfunc_all[pyfunc_idx]
    return ret_all

  def constraints_are_satisfied(self, point):
    """ Checks if the constraints are satisfied. """
    if hasattr(self, 'domain_constraints') and self.domain_constraints is not None:
      raw_point = self.get_raw_point(point)
      name_to_pt_dict = {k:v for (k, v) in zip(self.raw_name_ordering, raw_point)}
      ret_all = self._evaluate_all_constraints(raw_point, name_to_pt_dict)
      for idx, elem in enumerate(ret_all):
        if not isinstance(elem, (bool, np.bool, np.bool_)):
          raise ValueError(
            'Constraint %d:%s (%s) returned %s. It should return type bool.'%(idx,
            self.domain_constraints[idx][0], self.domain_constraints[idx][1], str(elem)))
      return all(ret_all)
    else:
      return True

  def members_are_equal(self, point_1, point_2):
    """ Compares two members and returns True if they are the same. """
    for i, dom in enumerate(self.list_of_domains):
      if not dom.members_are_equal(point_1[i], point_2[i]):
        return False
    return True

  def compute_distance(self, point_1, point_2):
    """ Computes the distance between point_1 and point_2. """
    return sum([dom.compute_distance(elem_1, elem_2) for (elem_1, elem_2, dom) in
                zip(point_1, point_2, self.list_of_domains)])

  def __str__(self):
    """ Returns a string representation of the domain. """
    list_of_domains_str = ', '.join([str(dom) for dom in self.list_of_domains])
    ret1 = 'CartProd(N=%d,d=%d)::[%s]'%(self.num_domains, self.dim,
                                        list_of_domains_str)
    if self.has_constraints():
      constraints_as_list_of_strs = ['%s: %s'%(elem[0], elem[1]) for elem in
                                     self.domain_constraints]
      constraints_as_str = ', '.join(constraints_as_list_of_strs)
      ret2 = ',  Constraints:: %s'%(constraints_as_str)
    else:
      ret2 = ''
    return ret1 + ret2


# Utilities we will need for the above ------------------------------------------
def is_within_bounds(bounds, point):
  """ Returns true if point is within bounds. point is a d-array and bounds is a
      dx2 array. bounds is expected to be an np.array object.
  """
  point = np.array(point)
  if point.shape != (bounds.shape[0],):
    return False
  above_lb = np.all((point - bounds[:, 0] >= 0))
  below_ub = np.all((bounds[:, 1] - point >= 0))
  return above_lb * below_ub

def _check_if_valid_euc_int_bounds(bounds):
  """ Checks if the bounds are valid. """
  for bd in bounds:
    if bd[0] > bd[1]:
      raise ValueError('Given bound %s is not in [lower_bound, upper_bound] format.'%(
                        bd))

def _get_bounds_as_str(bounds):
  """ returns a string representation of bounds. """
  bounds_list = [list(b) for b in bounds]
  return str(bounds_list)

def all_items_are_numeric(list_of_items):
  """ Returns true if all items in the list are numeric. """
  for elem in list_of_items:
    if not isinstance(elem, Number):
      return False
  return True

def all_lists_of_items_are_numeric(list_of_list_of_items):
  """ Returns true if all lists in list_of_list_of_items are numeric. """
  for elem in list_of_list_of_items:
    if not all_items_are_numeric(elem):
      return False
  return True

def discrete_numeric_element_is_in_list(elem, list_of_num_elements, tol=1e-8):
  """ Returns True if elem is in list_of_num_elements. Writing this separately due to
      precision issues with Python.
  """
  if not isinstance(elem, Number):
    return False
  # Iterate through the list and check if the element exists within tolerance.
  for list_elem in list_of_num_elements:
    if abs(elem - list_elem) < tol:
      return True
  return False




"""
  Utilities for plotting.
  -- syiblet@andrew.cmu.edu
  -- kvysyara@andrew.cmu.edu
  -- kandasamy@cs.cmu.edu
"""

from __future__ import division

# pylint: disable=invalid-name
# pylint: disable=redefined-builtin
# pylint: disable=too-many-locals
# pylint: disable=too-many-branches

import pickle
import argparse
import matplotlib.pyplot as plt
import matplotlib
from scipy.io import loadmat
import numpy as np

matplotlib.rcParams['mathtext.fontset'] = 'custom'
matplotlib.rcParams['mathtext.rm'] = 'Bitstream Vera Sans'
matplotlib.rcParams['mathtext.it'] = 'Bitstream Vera Sans:italic'
matplotlib.rcParams['mathtext.bf'] = 'Bitstream Vera Sans:bold'
matplotlib.rcParams['mathtext.fontset'] = 'stix'
matplotlib.rcParams['font.family'] = 'STIXGeneral'

NUM_GRID_PTS = 100
NUM_ERR_BARS = 10
LINE_WIDTH = 2
MARKER_SIZE = 6
AXIS_FONT_SIZE = 13
TITLE_FONT_SIZE = 18


def rgba(red, green, blue, a):
  '''rgba: generates matplotlib compatible rgba values from html-style rgba values
  '''
  return (red / 255.0, green / 255.0, blue / 255.0, a)

def transparent(red, green, blue, _, opacity=0.5):
  '''transparent: converts a rgba color to a transparent opacity
  '''
  return (red, green, blue, opacity)

def to_hex(hexstring):
  '''to_hex: generates matplotlib-compatible rgba values from html-style hex colours
  '''
  if hexstring[0] == '#':
    hexstring = hexstring[1:]
  red = int(hexstring[:2], 16)
  green = int(hexstring[2:4], 16)
  blue = int(hexstring[4:], 16)
  return rgba(red, green, blue, 1.0)

COLOURS = {
  'red': to_hex('#FF4136'),
  'blue': to_hex('#0074D9'),
  'cyan': to_hex('#00FFFF'),
  'magenta': to_hex('#FF00FF'),
  'green': to_hex('#2ECC40'),
  'orange': to_hex('#FF851B'),
  'lime': to_hex('#01FF70'),
  'fuchsia': to_hex('#F012BE'),
  'purple': to_hex('#B10DC9'),
  'olive': to_hex('#3D9970'),
  'teal': to_hex('#39CCCC'),
  'aqua': to_hex('#7FDBFF'),
  'maroon': to_hex('#85144B'),
  'yellow': to_hex('#FFDC00'),
  'navy': to_hex('#001F3F'),
  'white': to_hex('#FFFFFF'),
  'silver': to_hex('#DDDDDD'),
  'grey': to_hex('#AAAAAA'),
  'black': to_hex('#111111'),
}


def generate_legend_marker_colour_orders(plot_order, method_legend_colour_marker_dict):
  """ Generates a list of legends, colours, and markers based off the dictionary. """
  plot_legends = [method_legend_colour_marker_dict[meth]['legend'] for meth in plot_order]
  plot_colours = [method_legend_colour_marker_dict[meth]['colour'] for meth in plot_order]
  plot_markers = [method_legend_colour_marker_dict[meth]['marker'] for meth in plot_order]
  plot_linestyles = [method_legend_colour_marker_dict[meth]['linestyle']
                     for meth in plot_order]
  return plot_legends, plot_colours, plot_markers, plot_linestyles


def gen_curves(
    plot_order,
    results,
    x_label,
    y_label,
    plot_legends,
    plot_colours,
    plot_markers,
    plot_linestyles,
    options,
    x_bounds=None,
    y_bounds=None,
    outlier_frac=0.1,
    log_y=False,
    log_x=False,
    plot_title=None,
    study_name=None,
    num_workers=None,
    time_distro_str=None,
    fill_error=False,
    err_bar_freq=5,
    plot_type='semilogy',
    to_plot_legend=True,
    true_maxval=None,
):
  # pylint: disable=too-many-arguments
  # pylint: disable=too-many-branches
  # pylint: disable=too-many-statements
  # pylint: disable=unused-argument
  # pylint: disable=unused-variable
  """ Plots the curves given the experiment results.
  """

  num_methods, num_experiments = results['curr_opt_vals'].shape
  if true_maxval is None:
    true_maxval = np.asscalar(results['true_maxval'])

  if true_maxval is None or not np.isfinite(true_maxval):
    to_plot_regret = False
  else:
    to_plot_regret = True
    plot_type = 'semilogy'
    y_label = 'Regret'

  methods = [str(method).strip() for method in results['methods']]
  # Exclude incomplete experiments if present in any of the methods
  for i in range(num_methods):
    for vals in results['curr_opt_vals'][i, num_experiments::-1]:
      if isinstance(vals, str) or vals.dtype == np.dtype('<U1'):
        num_experiments = num_experiments - 1

  if x_bounds is None or x_bounds == []:
    x_bounds = [0.0, np.asscalar(results['max_capital'])]

  if log_x:
    grid_pts = np.logspace(
        np.log10(x_bounds[0]),
        np.log10(x_bounds[1]),
        num=NUM_GRID_PTS
    )
  else:
    grid_pts = np.linspace(
        x_bounds[0],
        x_bounds[1],
        num=NUM_GRID_PTS
    )

  err_bar_idx_half_gap = 0.5 * NUM_GRID_PTS / NUM_ERR_BARS
  err_bar_idxs = np.round(np.linspace(
      err_bar_idx_half_gap,
      NUM_GRID_PTS - err_bar_idx_half_gap,
      num=NUM_ERR_BARS
  )).astype(np.int)

  unordered_plot_means = np.zeros((num_methods, NUM_GRID_PTS))
  unordered_plot_stds = np.zeros((num_methods, NUM_GRID_PTS))
  init_opt_vals = None

  for i in range(num_methods):
    meth_curr_opt_vals = results['curr_true_opt_vals'][i, :]
    meth_costs = results['query_eval_times'][i, :]
    cum_costs = results['query_receive_times'][i, :]
    meth_plot_mean, meth_plot_std = get_plot_info(
        meth_curr_opt_vals,
        cum_costs,
        meth_costs,
        grid_pts,
        outlier_frac,
        init_opt_vals,
        num_experiments
    )

    if to_plot_regret:
      unordered_plot_means[i, :] = true_maxval - meth_plot_mean
    else:
      unordered_plot_means[i, :] = meth_plot_mean
    unordered_plot_stds[i, :] = meth_plot_std

  # re-order plot_means
  plot_means = np.zeros((num_methods, NUM_GRID_PTS))
  plot_stds = np.zeros((num_methods, NUM_GRID_PTS))
  for plot_idx, method in enumerate(plot_order):
    saved_order = methods.index(method)
    plot_means[plot_idx, :] = unordered_plot_means[saved_order, :]
    plot_stds[plot_idx, :] = unordered_plot_stds[saved_order, :]

  # Print out some statistics about plot_means
  all_plot_vals = plot_means.flatten()
  all_plot_vals = all_plot_vals[np.isfinite(all_plot_vals)]
  percentiles = [0.001, 0.1, 0.5, 0.9, 0.999]
  percentile_vals = np.percentile(all_plot_vals, percentiles)
  print_percentile_list = ['%0.3f:%0.4f'%(p, v) for (p, v) in
                           zip(percentiles, percentile_vals)]
  print('Percentiles:: %s.'%(',  '.join(print_percentile_list)))

  err_bar_pts = grid_pts[err_bar_idxs]
  err_bar_means = plot_means[:, err_bar_idxs]
  err_bar_stds = plot_stds[:, err_bar_idxs]

  if plot_type == 'plot':
    plot_func = plt.plot
  elif plot_type == 'loglog':
    plot_func = plt.loglog
  elif plot_type == 'semilogy':
    plot_func = plt.semilogy
  elif plot_type == 'semilogx':
    plot_func = plt.semilogx
  else:
    raise ValueError('Unknown plot function.')

  first_lines_for_legend = []

  # First the bars for the legend
#   for i, method in enumerate(methods):
  for plot_idx, method in enumerate(plot_order):
    # First plot the error bars
    curr_leg_line, = plot_func(
                       err_bar_pts,
                       err_bar_means[plot_idx],
                       marker=plot_markers[plot_idx],
                       color=plot_colours[plot_idx],
                       label=plot_legends[plot_idx],
                       linewidth=LINE_WIDTH,
                       linestyle=plot_linestyles[plot_idx],
                       markersize=MARKER_SIZE,
                       )
    first_lines_for_legend.append(curr_leg_line)
  if to_plot_legend:
    plt.legend(loc=options.legend_location, fontsize='large')
  # Now plot the whole curve
  for plot_idx, method in enumerate(plot_order):
    plot_func(
          grid_pts,
          plot_means[plot_idx, :],
          marker=',',
          color=plot_colours[plot_idx],
          linestyle=plot_linestyles[plot_idx],
          linewidth=LINE_WIDTH,
          )
  # Now do the error bars
  for plot_idx, method in enumerate(plot_order):
    if not fill_error:
      plt.errorbar(
          err_bar_pts,
          err_bar_means[plot_idx],
          err_bar_stds[plot_idx],
          marker=plot_markers[plot_idx],
          color=plot_colours[plot_idx],
          linewidth=LINE_WIDTH,
          markersize=MARKER_SIZE,
          linestyle='',
      )
    else:
      plt.fill_between(
          grid_pts,
          plot_means[i, :] - plot_stds[i, :],
          plot_means[i, :] + plot_stds[i, :],
          color=transparent(*plot_colours[plot_idx], opacity=0.3),
      )

  if plot_title is not None:
    plt.title(plot_title, fontsize=TITLE_FONT_SIZE)
#   if to_plot_regret:
#     plt.title('{}; workers={}; {}; Exps={}; Max Val={}'.format(study_name, num_workers,
#                       time_distro_str, num_experiments, true_maxval).replace('_', '-'))
#   else:
#     plt.title('{}; Workers={}; {}; Exps={}'.format(study_name, num_workers,
#                       time_distro_str, num_experiments).replace('_', '-'))
  plt.xlabel(x_label, fontsize=AXIS_FONT_SIZE)
  plt.ylabel(y_label, fontsize=AXIS_FONT_SIZE)
  if y_bounds is not None:
    plt.ylim(y_bounds)
  # Remove duplicate lines
  for leg_line in first_lines_for_legend:
    leg_line.remove()
  plt.draw()
  plt.show()


# Utilities to read and process data -----------------------------------------------
def get_plot_info(
    meth_curr_opt_vals,
    cum_costs,
    meth_costs,
    grid_pts,
    outlier_frac,
    init_opt_vals,
    num_experiments
):
  """ Generates means and standard deviation for the method's output
  """
  outlier_low_idx = int(max(np.round(outlier_frac * num_experiments), 0))
  outlier_high_idx = min(
      num_experiments,
      int(num_experiments  - np.rint(outlier_frac * num_experiments))
  )
  # Manage potential outliers
  inlier_idx = 0
  if num_experiments > 1:
    inlier_idx = np.arange(outlier_low_idx, outlier_high_idx)
  num_grid_pts = len(grid_pts)
  grid_vals = np.zeros((num_experiments, num_grid_pts))
  # Iterate through each experiment
  for exp_iter in range(num_experiments):
    if cum_costs is None:
      curr_cum_costs = np.cumsum(meth_costs[exp_iter])
    else:
      curr_cum_costs = cum_costs[exp_iter]
    if init_opt_vals is not None:
      opt_vals = np.concatenate((np.array([init_opt_vals[exp_iter]]),
                                 np.squeeze(meth_curr_opt_vals[exp_iter])), axis=0)
      curr_cum_costs = np.concatenate((np.array([0]), np.squeeze(curr_cum_costs)),
                                      axis=0)
    else:
      opt_vals = meth_curr_opt_vals[exp_iter]
    interp = np.interp(grid_pts, curr_cum_costs.flatten(), opt_vals.flatten())
    grid_vals[exp_iter, :] = np.fmax.accumulate(interp)
  sorted_grid_vals = np.sort(grid_vals, axis=0)
  inlier_grid_vals = sorted_grid_vals[inlier_idx, :]
  if num_experiments == 1:
    inlier_grid_vals = np.expand_dims(inlier_grid_vals, axis=0)
  # An internal function to get mean and std
  def mean_and_std(arr1d):
    """ Returns mean and standard deviation."""
    finite_arr1d = arr1d[np.isfinite(arr1d)]
    if finite_arr1d.size / arr1d.size >= 0.4:
      return np.array([np.mean(finite_arr1d), np.std(finite_arr1d) / np.sqrt(arr1d.size)])
    return np.array([np.NaN] * 2)
  # Return
  res = np.apply_along_axis(mean_and_std, 0, inlier_grid_vals)
  return (res[0, :], res[1, :])



# Utilities to read results from a file --------------------------------------------
def read_results(file_path):
  """reads experiment result data from a '.m' file
  :file_path: the path to the file
  :returns: a dataframe object with all the various pieces of data

  """
  if file_path.endswith('.mat'):
    results = loadmat(file_path)
  elif file_path.endswith('.p'):
    with open(file_path, 'rb') as pickleF:
      res = pickle.load(pickleF)
      pickleF.close()

    results = {}
    for key in list(res.keys()):
      if not hasattr(res[key], '__len__'):
        results[key] = np.array(res[key])
      elif isinstance(res[key], str):
        results[key] = np.array(res[key])
      elif isinstance(res[key], list):
        results[key] = np.array(res[key])
      elif isinstance(res[key], np.ndarray):
        val = np.zeros(res[key].shape, dtype=res[key].dtype)
        for idx, x in np.ndenumerate(res[key]):
          if isinstance(x, list):
            val[idx] = np.array(x)
          else:
            val[idx] = x
        results[key] = val
      else:
        results[key] = res[key]
  else:
    raise ValueError('Wrong file format. It has to be either mat or pickle file')
  return results

# Utilities for loading results and plotting ------------------------------------------
def load_results(file_paths):
  """ Concatenates the results from multiple files. """
  results = read_results(file_paths[0])
  ignore = ['study_name', 'num_workers', 'time_distro_str', 'max_capital',
            'true_maxval', 'num_jobs_per_worker', 'domain_type', 'true_argmax']
  for i in range(1, len(file_paths)):
    results_i = read_results(file_paths[i])
    diff_methods = False
    if len(results_i['methods']) != len(results['methods']):
      diff_methods = True
    elif not (results_i['methods'] == results['methods']).all():
      diff_methods = True
    for key in list(results_i.keys()):
      if key in ignore:
        continue
      elif isinstance(results_i[key], np.ndarray):
        if results_i[key].ndim == 2:
          if not diff_methods:
            results[key] = np.concatenate((results[key], results_i[key]), axis=1)
          else:
            if results[key].ndim != 2:
              results[key] = np.expand_dims(results[key], axis=0)
            _, n1 = results_i[key].shape
            _, n = results[key].shape
            if n1 < n:
              n = n1
            results[key] = np.concatenate((results[key][:, :n], results_i[key][:, :n]),
                                          axis=0)
        elif results_i[key].ndim == 0:
          continue
        else:
          results[key] = np.concatenate((results[key], results_i[key]))
  return results

def get_file_paths(fname):
  """ Read file paths from the file. """
  with open(fname) as f:
    file_paths = f.readlines()
  file_paths = [path.rstrip() for path in file_paths if path.rstrip()]
  return file_paths


def get_plot_options():
  """ Given a list of options, this reads them from the command line and returns
      a namespace with the values.
  """
  parser = argparse.ArgumentParser(description='Plotting.')
  parser.add_argument('--file', default='', help='File path of single plot file.')
  parser.add_argument('--filelist', default='', help='File name containing file paths.')
  parser.add_argument('--type', default='semilogy', help='Type of plot. Default is ' +
                      'semilogy, other options are plot, loglog, semilogx.')
  parser.add_argument('--title', help='Title of plot.')
  options = parser.parse_args()

  return options


def plot_results(results, plot_order, method_legend_colour_marker_dict, x_label,
                 y_label, x_bounds=None, y_bounds=None, to_plot_legend=True,
                 true_maxval=None, outlier_frac=0.0, plot_title=None, options=None):
  # pylint: disable=too-many-arguments
  """ Plots the results using Matplotlib. """
  if options is None:
    options = get_plot_options()
  if plot_title == None:
    plot_title = options.title
  if x_bounds is None:
    x_bounds = []
  # Get order of legends, colours, etc.
  plot_legends, plot_colours, plot_markers, plot_linestyles = \
    generate_legend_marker_colour_orders(plot_order, method_legend_colour_marker_dict)
  return gen_curves(
      plot_order,
      results,
      x_label,
      y_label,
      plot_legends,
      plot_colours,
      plot_markers,
      plot_linestyles,
      options,
      x_bounds=x_bounds,
      y_bounds=y_bounds,
      outlier_frac=outlier_frac,
      log_y=True,
      plot_title=plot_title,
      study_name=str(np.asscalar(results['study_name'])),
      num_workers=str(np.asscalar(results['num_workers'])),
      time_distro_str=str(np.asscalar(results['time_distro_str'])),
      plot_type=options.type,
      to_plot_legend=to_plot_legend,
      true_maxval=true_maxval,
  )

